#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  configfile.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import yaml
from easydict import EasyDict
from jsonschema import validate
from .schema import SCHEMA

LOG_FILTER_TEMPLATE = '''
    {{%- if data.level <= {loglevel} -%}}
        pass
    {{%- else -%}}
        nomatch
    {{%- endif -%}}'''

LOG_COLOR_TEMPLATE = '''
    {%- if data.level == 0 -%}
        \x1B[0;35m
    {%- elif data.level == 1 -%}
        \x1B[1;35m
    {%- elif data.level == 2 -%}
        \x1B[0;31m
    {%- elif data.level == 3 -%}
        \x1B[1;31m
    {%- elif data.level == 4 -%}
        \x1B[1;33m
    {%- elif data.level == 5 -%}
        \x1B[1;30m
    {%- elif data.level == 6 -%}
        \x1B[1;37m
    {%- else -%}
        \x1B[1;37m
    {%- endif -%}
    {{strftime(epoch(), "YYYY-MM-DDTHH:mm:ss.SSSSZZ")}} {{data.identification}}[{{data.pid}}] {{data.txt_level}} {{data.module}}: {{data.message}}\x1B[0m'''


class ConfigFile(object):
    '''
    Generates a wishbone.router configuration object used to initialize a
    wishbone router object.

    This config generator has some tailered functionality which makes it
    suitable when bootstrapping from CLI.

    It does following automatic configurations:

    - Initializes a ``wishbone.module.flow.funnel`` module called ``_logs``
      which is connected to the ``_logs`` queue of all modules except if
      this module has already been connected in the bootstrap file.

    - Initializes a ``wishbone.module.flow.funnel`` module called ``_metrics``
      which is connected to the ``_metrics`` queue of all modules except if
      this module has already been connected in the bootstrap file.  The
      ``_metrics`` modules is by default not connected to any other
      modules.  The effect of this is that all metrics are dropped unless
      the user connects a module for furhter processing the metrics.

    - Adds a ``wishbone.module.flow.queueselect`` module called
      ``_logs_filter`` responsible for dropping logs which log level do
      not correspond to the define ``--log-level``

    - Adds either a ``wishbone.module.output.stdout`` called
      ``_logs_stdout`` module or ``wishbone.module.output.syslog`` module
      called ``_logs_syslog`` and connects this instance to
      ``_logs.outbox``.

    - Initializes the following template functions and makes them
      available to each initialized module::

        - strftime()
        - env()
        - epoch()
        - version()

    Args:
        filename (str): The filename of the configuration to load.
        logstyle (str): How logging should be setup. Possible options are: stdout and syslog.
        loglevel (str): The loglevel for the router to use to use.
        identification (str): A string which identifies the router instance
        colorize_stdout (bool): When True, colors each stdout printed logline using proper coloring.
    '''

    def __init__(self, filename, logstyle, loglevel=6, identification="wishbone", colorize_stdout=True):
        self.identification = identification
        self.colorize_stdout = colorize_stdout
        self.logstyle = logstyle
        self.loglevel = loglevel
        self.config = EasyDict({
            "template_functions": EasyDict({}),
            "modules": EasyDict({}),
            "module_functions": EasyDict({}),
            "protocols": EasyDict({}),
            "routingtable": []
        })
        self.__addLogFunnel()
        self.__addLogFilter()
        self.__addMetricFunnel()
        self.load(filename)

    def addModule(self, name, module, arguments={}, description="", functions={}, protocol=None, event=False):
        '''
        Adds a module to the configuration.

        Args:
            name (str): The module instance name
            module (str): The module name
            arguments (dict): The module variables
            description (str): A description of the module instance
            functions (dict): The module functions
            protocol (str): The protocol to apply to the module
            event (bool): Whether incoming or outgoing events need to be treated as full events.
        '''

        if name.startswith('_'):
            raise Exception("Module instance names cannot start with _.")

        self.__addModule(name, module, arguments, description, functions, protocol, event)

    def addTemplateFunction(self, name, function, arguments={}):
        '''Adds a template funtion to the configuration.

        Args:
            name (str): The name of the function instance
            function (str): The entry point name of the function
            arguments (dict): The arguments to initiate the function class.
        '''

        if name not in self.config["template_functions"]:
            self.config["template_functions"][name] = EasyDict({
                "function": function,
                "arguments": arguments
            })
        else:
            raise Exception("Lookup instance name '%s' is already taken." % (name))

    def addModuleFunction(self, name, function, arguments={}):
        '''
        Adds a module function to the configuration.

        Args:
            name (str): The name of the function instance
            function (str): The entry point name of the function
            arguments (dict): The arguments to initiate the function class.
        '''

        self.config["module_functions"][name] = EasyDict({"function": function, "arguments": arguments})

    def addProtocol(self, name, protocol, arguments={}):
        '''
        Adds a protocol to the configuration

        Args:
            name (str): The name of the protocol instance
            protcol (str): The entry point name of the protocol
            arguments (dict): The arguments to initiate the protocol class.
        '''

        self.config["protocols"][name] = EasyDict({"protocol": protocol, "arguments": arguments})

    def addConnection(self, source_module, source_queue, destination_module, destination_queue):
        '''
        Adds connections between module queues.

        Args:
            source_module (str): The source module instance name
            source_queue (str): The source module instance queue name
            destination_module (str): The destination instance name
            destination_queue (str): The destination instance queue name
        '''
        connected = self.__queueConnected(source_module, source_queue)

        if not connected:
            self.config["routingtable"].append(
                EasyDict({
                    "source_module": source_module,
                    "source_queue": source_queue,
                    "destination_module": destination_module,
                    "destination_queue": destination_queue
                })
            )
        else:
            raise Exception("Cannot connect '%s.%s' to '%s.%s'. Reason: %s." % (source_module, source_queue, destination_module, destination_queue, connected))

    def dump(self):
        '''
        Dumps the configuration as an ``EasyDict`` instance.
        '''

        return EasyDict(self.config)

    def load(self, filename):
        '''
        Loads a YAML bootstrap file.

        Args:
            filename (str): The filename to load.
        '''

        config = self.__load(filename)
        self.__validate(config)
        self.__validateRoutingTable(config)

        self.__addDefaultTemplateFunctions()

        if "template_functions" in config:
            for function in config["template_functions"]:
                self.addTemplateFunction(name=function, **config["template_functions"][function])

        if "module_functions" in config:
            for function in config["module_functions"]:
                self.addModuleFunction(name=function, **config["module_functions"][function])

        if "protocols" in config:
            for protocol in config["protocols"]:
                self.addProtocol(
                    name=protocol,
                    protocol=config["protocols"][protocol].get("protocol", None),
                    arguments=config["protocols"][protocol].get("arguments", {}),
                )

        for module in config["modules"]:
            self.addModule(name=module, **config["modules"][module])

        for route in config["routingtable"]:
            sm, sq, dm, dq = self.__splitRoute(route)
            self.addConnection(sm, sq, dm, dq)

        getattr(self, "_setupLogging%s" % (self.logstyle.upper()))()

    def __addDefaultTemplateFunctions(self):
        '''
        Adds template functions which should be available for each module.
        '''

        self.addTemplateFunction("strftime", "wishbone.function.template.strftime")
        self.addTemplateFunction("epoch", "wishbone.function.template.epoch")
        self.addTemplateFunction("env", "wishbone.function.template.environment")
        self.addTemplateFunction("version", "wishbone.function.template.version")

    def __addModule(self, name, module, arguments={}, description="", functions={}, protocol=None, event=False):

        if protocol is not None and protocol not in self.config.protocols:
            raise Exception("No protocol module defined with name '%s' for module instance '%s'" % (protocol, name))

        for queue, fs in functions.items():
            for function in fs:
                if function not in self.config.module_functions.keys():
                    raise Exception("No function defined with name '%s' for module instance '%s'." % (function, name))

        if name not in self.config["modules"]:
            self.config["modules"][name] = EasyDict({
                'description': description,
                'module': module,
                'arguments': arguments,
                'functions': functions,
                'protocol': protocol,
                'event': event})
            self.addConnection(name, "_logs", "_logs", "_%s" % (name))
            self.addConnection(name, "_metrics", "_metrics", "_%s" % (name))

        else:
            raise Exception("Module instance name '%s' is already taken." % (name))

    def __queueConnected(self, module, queue):

        for c in self.config["routingtable"]:
            if (c["source_module"] == module and c["source_queue"] == queue) or (c["destination_module"] == module and c["destination_queue"] == queue):
                return "Queue '%s.%s' is already connected to '%s.%s'" % (c["source_module"], c["source_queue"], c["destination_module"], c["destination_queue"])
        return False

    def __splitRoute(self, route):

        (source, destination) = route.split('->')
        (source_module, source_queue) = source.rstrip().lstrip().split('.')
        (destination_module, destination_queue) = destination.rstrip().lstrip().split('.')
        return source_module, source_queue, destination_module, destination_queue

    def __load(self, filename):
        '''Loads and returns the yaml bootstrap file.'''

        try:
            with open(filename, 'r') as f:
                config = yaml.load(f)
        except Exception as err:
            raise Exception("Failed to load bootstrap file.  Reason: %s" % (err))
        else:
            return config

    def __validate(self, config):

        try:
            validate(config, SCHEMA)
        except Exception as err:
            raise Exception("Failed to validate configuration file.  Reason: %s" % (err.message))

    def __validateRoutingTable(self, config):

        for route in config["routingtable"]:
            (left, right) = route.split("->")
            assert "." in left.lstrip().rstrip(), "routingtable rule \"%s\" does not have the right format. Missing a dot." % (route)
            assert "." in right.lstrip().rstrip(), "routingtable rule \"%s\" does not have the right format. Missing a dot." % (route)

    def __addLogFilter(self):

        self.__addModule(
            name="_logs_filter",
            module="wishbone.module.flow.queueselect",
            arguments={
                "templates": [
                    {"name": "log_filter",
                     "queue": LOG_FILTER_TEMPLATE.format(loglevel=self.loglevel)
                     }
                ]
            },
            description="Centralizes the logs of all modules.",
            functions={
            },
            protocol=None
        )

    def __addLogFunnel(self):

        self.__addModule(
            name="_logs",
            module="wishbone.module.flow.funnel",
            arguments={
            },
            description="Centralizes the logs of all modules.",
            functions={
            },
            protocol=None
        )

    def __addMetricFunnel(self):

        self.__addModule(
            name="_metrics",
            module="wishbone.module.flow.funnel",
            arguments={
            },
            description="Centralizes the metrics of all modules.",
            functions={
            },
            protocol=None
        )

    def _setupLoggingSTDOUT(self):

        if not self.__queueConnected("_logs", "outbox"):

            self.__addModule(
                name="_logs_format",
                module="wishbone.module.process.template",
                arguments={
                    "templates": {
                        "human_log": LOG_COLOR_TEMPLATE
                    }
                },
                description="Create a human readable log format.",
                functions={
                },
                protocol=None
            )

            self.addConnection("_logs", "outbox", "_logs_filter", "inbox")
            self.addConnection("_logs_filter", "pass", "_logs_format", "inbox")

            self.__addModule(
                name="_logs_stdout",
                module="wishbone.module.output.stdout",
                arguments={
                    "colorize": self.colorize_stdout,
                    "selection": "human_log"
                },
                description="Prints all incoming logs to STDOUT.",
                functions={
                },
                protocol=None
            )
            self.addConnection("_logs_format", "outbox", "_logs_stdout", "inbox")

    def _setupLoggingSYSLOG(self):

        if not self.__queueConnected("_logs", "outbox"):

            self.__addModule(
                name="_logs_syslog",
                module="wishbone.module.output.syslog",
                arguments={
                    "ident": self.identification,
                    "payload": "module({{data.module}}): {{data.message}}"
                },
                description="Writes all incoming messages to syslog.",
                functions={
                },
                protocol=None
            )

            self.addConnection("_logs", "outbox", "_logs_filter", "inbox")
            self.addConnection("_logs_filter", "pass", "_logs_syslog", "inbox")



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  template.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from jinja2 import Environment, FileSystemLoader
from wishbone.module import ProcessModule
from os.path import isabs


class Template(ProcessModule):

    '''**Renders Jinja2 templates.**

    Renders Jinja2 templates using the event content as context. Templates can
    be read from a file or directly from the bootstrap file.


    Parameters::

        - filename(str)(None)*
           |  The absolute template filename to load.
           |  Can be a dynamic lookup. Assing None
           |  (default) to disable.

        - destination(str)(data)*
           |  The key to store the rendered file template.


        - templates(dict)({})*
           |  A dict of templates to render. The dict keys are the fields to
           |  assign the rendered string to.


    Queues::

        - inbox
           |  Incoming events

        - outbox
           |  Outgoing events

    '''

    def __init__(self, actor_config, filename=None, destination="data", templates={}):
        ProcessModule.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

        self.__template_loaders = {}

        if self.kwargs.filename is not None:
            self.file_template = Environment(loader=FileSystemLoader(self.kwargs.filename))

    def consume(self, event):

        for key, result in event.kwargs.templates.items():
            event.set(result, key)

        if event.kwargs.filename is not None:

            if isabs(event.kwargs.filename):
                data = self.renderFile(event.kwargs.filename, event.dump())
                event.set(data, event.kwargs.destination)
            else:
                self.logging.error("%s is not an absolute path. Skipped" % (event.kwargs.filename))

        self.submit(event, "outbox")

    def renderFile(self, filename, data):

        # TODO(smetj): calling thie private function is ugly, rewrite this part.
        return self._Actor__renderKwargs.env_template.get_template(filename).render(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  event.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import time
from wishbone.error import BulkFull, InvalidData, TTLExpired
from uuid import uuid4
from jinja2 import Template
from copy import deepcopy
from scalpl import Cut
from easydict import EasyDict


EVENT_RESERVED = ["timestamp", "data", "tmp", "errors", "uuid", "uuid_previous", "cloned", "bulk", "ttl", "tags"]


def extractBulkItemValues(event, selection):
    '''Yields a field from all events in the bulk event.

    Args:

        event (wishbone.event.Event): The wishbone event in bulk mode.
        selection (str): The field to extract form each event in the bulk.


    Yields:

        str/int/float/dict/list: The value of the event
    '''

    if event.isBulk():
        for e in event.data["data"]:
            yield Event().slurp(e).get(selection)


def extractBulkItems(event):
    '''Yields all the events from a bulk event.

    Args:

        event (wishbone.event.Event): The wishbone event in bulk mode.

    Yields:

        `wishbone.event.Event`: The value of the event
    '''

    if event.isBulk():
        for e in event.data["data"]:
            yield Event().slurp(e)


class Event(object):

    '''The Wishbone event object.

    A class object containing the event data being passed from one Wishbone
    module to the other.

    The keyformat used is the one handled by the ``Scalpl`` module
    (https://pypi.python.org/pypi/scalpl/).

    Args:

        data (dict/list/string/int/float): The data to assign to the ``data`` field.
        ttl (int): The TTL value for the event.
        bulk (bool): Initialize the event as a bulk event
        bulk_size (int): The number of events the bulk can hold.

    Attributes:

        data (dict): A dict containing the event data structure.
        bulk_size (int): The max allowed bulk size.
    '''

    def __init__(self, data=None, ttl=254, bulk=False, bulk_size=100):

        self.data = Cut({
            "cloned": False,
            "bulk": bulk,
            "data": None,
            "errors": {
            },
            "tags": [],
            "timestamp": time.time(),
            "tmp": {
            },
            "ttl": ttl,

            "uuid_previous": [
            ],
        })

        if bulk:
            self.data["data"] = []
        else:
            self.set(data)
        self.data["uuid"] = str(uuid4())
        self.bulk_size = bulk_size

    def appendBulk(self, event):
        '''Appends an event to this bulk event.

        Args:

            event(wishbone.event.Event): The event to add to the bulk instance

        Raises:

            InvalidData: Either the event is not of type Bulk or ``event`` is
                         not an wishbone.event.Event instance.
        '''

        if self.data["bulk"]:
            if isinstance(event, Event):
                if len(self.data["data"]) == self.bulk_size:
                    raise BulkFull("The bulk event already contains '%s' events." % (self.bulk_size))

                self.data["data"].append(event.dump())
            else:
                raise InvalidData("'event' should be of type wishbone.event.Event.")
        else:
            raise InvalidData("This instance is not initialized as a bulk event.")

    def clone(self):
        '''Returns a cloned version of the event.

        Returns:

            class: A ````wishbone.event.Event```` instance


        '''
        e = deepcopy(self)

        if "uuid_previous" in e.data:
            e.data["uuid_previous"].append(
                e.data["uuid"]
            )
        else:
            e.data["uuid_previous"] = [
                e.data["uuid"]
            ]
        e.data["uuid"] = str(uuid4())
        e.data["timestamp"] = time.time()
        e.data["cloned"] = True

        return e

    def copy(self, source, destination):
        '''Copies the source key to the destination key.

        Args:

            source (str): The name of the source key.
            destination (str): The name of the destination key.
        '''

        self.set(
            deepcopy(
                self.get(
                    source
                )
            ),
            destination
        )

    def decrementTTL(self):
        '''Decrements the TTL value.

        Raises:

            TTLExpired: When TTL has reached 0.
        '''

        self.data["ttl"] -= 1

        if self.data["ttl"] <= 0:
            raise TTLExpired("Event TTL expired in transit.")

    def delete(self, key=None):
        '''Deletes a key.

        Args:

            key (str): The key to delete

        Raises:

            Exception: Deleting the root of a reserved keyword such as ``data`` or ``tags``.
            KeyError: When a non-existing key is referred to.
        '''

        s = key.split('.')
        if s[0] in EVENT_RESERVED and len(s) == 1:
            raise Exception("Cannot delete root of reserved keyword '%s'." % (key))

        del(self.data[key])

    def dump(self):
        '''
        Dumps the complete event.
        This complete event is also called a ``native event``

        Returns:

            dict: The content of the event.
        '''

        d = deepcopy(self)
        d.data["timestamp"] = float(d.data["timestamp"])
        return d.data.data

    def get(self, key="data"):
        '''Returns the value of ``key``.

        ``key`` must be in ``Scalpl`` format.

        Args:

            key (str): The name of the key to read.

        Returns:

            str/int/float/dict/list: The value of the key

        Raises:

            KeyError: The provided key does not exist.
        '''
        if key in [None, "", "."]:
            return self.data
        else:
            try:
                return self.data[key]
            except Exception as err:
                raise KeyError(key)

    def has(self, key="data"):
        '''Returns a bool indicating the event has ``key``

        ``key`` must be in ``Scalpl`` format.

        Args:

            key (str): The name of the key to check

        Returns:

            bool: True if the key is there otherwise false

        Raises:

            KeyError: The provided key does not exist
        '''

        try:
            self.data[key]
        except KeyError:
            return False
        else:
            return True

    def isBulk(self):
        '''Tells whether event is ``bulk`` or not.


        Returns:

            bool: True if the event is bulk

        '''

        return self.data["bulk"]

    def merge(self, value, key="data"):
        '''
        Merges value into ``key``.
        Value types should be mergeable otherwise an error is returned.

        Args:
            value (dict,list): The value to merge
            key (str): The key to merge into

        Raises:
            InvalidData: Types are not mergeable
        '''

        try:
            if isinstance(value, list):
                self.data[key] += value
            elif isinstance(value, dict):
                self.data[key].update(value)
            else:
                raise InvalidData("Source and destination are incompatible to merge")
        except Exception:
            raise InvalidData("Source and destination are incompatible to merge")

    def render(self, template, env_template=None):
        '''Returns a formatted string using the provided template and key

        Args:

            template (str): A string representing the Jinja2 template.
            env_template (``jinja2.Environment``): Used to render template strings from.
                If not set, then template rendering happens without environment.

        Returns:

            str: The rendered string

        Raises:

            InvalidData: An invalid jinja2 template has been provided
        '''

        try:
            if env_template is None:
                return Template(template).render(**self.data)
            else:
                return env_template.from_string(template).render(**self.data)
        except Exception as err:
            raise InvalidData("Failed to render template. Reason: %s" % (err))

    def renderField(self, field_name, env_template=None):
        '''
        Expects ``field_name`` to contain a template, renders it and replaces
        its content with the result. If the field ``field_name`` contains
        anything else than a string then it's silently ignored.

        Args:

            field_name (str): A string defining the field to handle.

        Returns:

            None

        Raises:

            InvalidData: An invalid jinja2 template has been provided
        '''

        try:
            raw_value = self.get(field_name)
            rendered_value = self.__renderDataStructure(raw_value, env_template)
            self.set(rendered_value, field_name)
        except Exception as err:
            raise InvalidData("Failed to render template. Reason: %s" % (err))

    def renderKwargs(self, template_kwargs):
        '''
        Renders all the templates found in ``template_kwargs`` and sets
        self.kwarg, a version of the current module's kwargs relate to this
        events' content

        Args:

            template_kwargs (dict): A dict of the modules kwargs optoinally
                                    containing Template instances.
        '''

        def recurse(data):

            if isinstance(data, Template):
                try:
                    return data.render(**self.dump())
                except Exception as err:
                    return "#error: %s#" % (err)
            elif isinstance(data, dict):
                result = {}
                for key, value in data.items():
                    result[key] = recurse(value)
                return EasyDict(result)
            elif isinstance(data, list):
                result = []
                for value in data:
                    result.append(recurse(value))
                return result
            else:
                return data

        self.kwargs = EasyDict(
            recurse(
                template_kwargs
            )
        )

    def set(self, value, key="data"):
        '''Sets the value of ``key``.

        ``key`` must be in ``Scalpl`` format.

        Args:

            value (str, int, float, dict, list): The value to assign.
            key (str): The key to store the value
        '''

        self.data[key] = value

    def slurp(self, data):
        '''
        Create an event object from a ``native event`` dict exported by
        ``dump()``

        The timestamp field will be reset to the time this method has been
        called.


        Args:

            data (dict): The dict object containing the complete event.


        Returns:

            wishbone.event.Event: A Wishbone event instance.


        Raises:

            InvalidData:  ``data`` does not contain valid fields to build
                                  an event
        '''
        try:
            assert isinstance(data, (dict, Cut)), "event.slurp() expects a dict."
            for item in [
                ("timestamp", float),
                ("data", None),
                ("tmp", dict),
                ("errors", dict),
                ("uuid", str),
                ("uuid_previous", list),
                ("cloned", bool),
                ("bulk", bool),
                ("ttl", int),
                ("tags", list)
            ]:
                assert item[0] in data, "%s is missing" % (item[0])
                if item[1] is not None:
                    assert isinstance(data[item[0]], item[1]), "%s type '%s' is not valid." % (item[0], item[1])
        except AssertionError as err:
            raise InvalidData("The incoming data could not be used to construct an event.  Reason: '%s'." % err)
        else:
            self.data = Cut(data)
            self.data["timestamp"] = time.time()

        return(self)

    raw = dump

    def __renderDataStructure(self, datastructure, env_template):

        def recurse(data):

            if isinstance(data, str):
                if '{{' in data and '}}'in data:
                    try:
                        if env_template is None:
                            return Template(data).render(**self.dump())
                        else:
                            return env_template.from_string(data).render(**self.dump())
                    except Exception as err:
                        return "#error: %s#" % (err)
                else:
                    return data
            elif isinstance(data, dict):
                result = {}
                for key, value in data.items():
                    result[key] = recurse(value)
                return EasyDict(result)
            elif isinstance(data, list):
                result = []
                for value in data:
                    result.append(recurse(value))
                return result
            else:
                return data

        return recurse(datastructure)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  funnel.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import FlowModule


class Funnel(FlowModule):

    '''**Funnel multiple incoming queues to one outgoing queue.**

    Funnel multiple incoming queues to one outgoing queue.

    Parameters::

        n/a

    Queues::

        - outbox:
           |  Outgoing messages

    '''

    def __init__(self, actor_config):

        FlowModule.__init__(self, actor_config)
        self.pool.createQueue("outbox")

    def preHook(self):

        for queue in self.pool.listQueues(default=False, names=True):
            if queue != "outbox":
                self.registerConsumer(self.consume, queue)

    def consume(self, event):

        self.submit(event, "outbox")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.error import ProtocolError
import types
from scalpl import Cut


class Decode(object):

    handle_buffer_size = False
    current_size = 0

    def flush(self):

        raise Exception("This decoder has no flush() method implemented.")

    def getBufferSize(self):

        return 0

    def handler(self, data):

        if not self.handle_buffer_size and self.getBufferSize() >= self.buffer_size:
            raise ProtocolError("Buffer full")
        try:
            if data is None:
                return self.flush()
            elif isinstance(data, bytes):
                return self.handleBytes(data)
            elif isinstance(data, str):
                return self.handleString(data)
            elif isinstance(data, int):
                return self.handleInt(data)
            elif isinstance(data, float):
                return self.handleFloat(data)
            elif isinstance(data, dict):
                return self.handleDict(data)
            elif isinstance(data, Cut):
                return self.handleDict(
                    dict(data)
                )
            elif isinstance(data, list):
                return self.handleList(data)
            elif isinstance(data, types.GeneratorType):
                return self.handleGenerator(data)
            elif hasattr(data, "readlines") and callable(data.readlines):
                return self.handleReadLinesMethod(data)
            else:
                raise ProtocolError("%s is not supported by this Decoder." % (type(data)))
        except Exception as err:
            raise ProtocolError(err)

    def handleBytes(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleDict(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleFloat(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleGenerator(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleInt(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleList(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleReadLinesMethod(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleString(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def handleUnicode(self, data):
        raise ProtocolError("%s is not supported by this Decoder." % (type(data)))

    def resetBuffer(self):

        raise Exception("Decoder has no resetBuffer() method.")

class Encode(object):

    def handler(self, data):
        try:
            if isinstance(data, bytes):
                return self.handleBytes(data)
            elif isinstance(data, str):
                return self.handleString(data)
            elif isinstance(data, int):
                return self.handleInt(data)
            elif isinstance(data, float):
                return self.handleFloat(data)
            elif isinstance(data, dict):
                return self.handleDict(data)
            elif isinstance(data, list):
                return self.handleList(data)
            elif isinstance(data, Cut):
                return self.handleCut(data)
            else:
                raise ProtocolError("%s is not supported by this Decoder." % (type(data)))
        except Exception as err:
            raise ProtocolError(err)

    def handleBytes(self, data):
        raise ProtocolError("%s is not supported by this Encoder." % (type(data)))

    def handleString(self, data):
        raise ProtocolError("%s is not supported by this Encoder." % (type(data)))

    def handleInt(self, data):
        raise ProtocolError("%s is not supported by this Encoder." % (type(data)))

    def handleFloat(self, data):
        raise ProtocolError("%s is not supported by this Encoder." % (type(data)))

    def handleUnicode(self, data):
        raise ProtocolError("%s is not supported by this Encoder." % (type(data)))

    def handleDict(self, data):
        raise ProtocolError("%s is not supported by this Encoder." % (type(data)))

    def handleList(self, data):
        raise ProtocolError("%s is not supported by this Encoder." % (type(data)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  wb_inotify.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import InputModule

# I know no other working way to actually monkey patch select
# when inotify is imported.
# If you know another way, please consider submitting a
# patch or merge request
# todo(smetj): Fix dirty monkey patch of inotify select
import sys
from gevent import monkey; monkey.patch_all()
sys.modules["select"] = sys.modules["gevent.select"]
sys.modules["select"].epoll = sys.modules["select"].poll

from inotify.adapters import Inotify
from inotify import constants

from gevent import sleep
import os
import fnmatch


class WBInotify(InputModule):

    '''**Monitors one or more paths for inotify events.**

    Monitors one or more paths for the defined inotify events.

    Inotify events can have following values:

        - IN_ACCESS
        - IN_ALL_EVENTS
        - IN_ATTRIB
        - IN_CLOEXEC
        - IN_CLOSE
        - IN_CLOSE_NOWRITE
        - IN_CLOSE_WRITE
        - IN_CREATE
        - IN_DELETE
        - IN_DELETE_SELF
        - IN_DONT_FOLLOW
        - IN_IGNORED
        - IN_ISDIR
        - IN_MASK_ADD
        - IN_MODIFY
        - IN_MOVE
        - IN_MOVED_FROM
        - IN_MOVED_TO
        - IN_MOVE_SELF
        - IN_NONBLOCK
        - IN_ONESHOT
        - IN_ONLYDIR
        - IN_OPEN
        - IN_Q_OVERFLOW
        - IN_UNMOUNT


    Outgoing events have following format::

        {"path": "/tmp/test", "inotify_type": "IN_ACCESS"}


    Parameters::

        - destination(str)(data)
           |  In which field to store the inotify event data.

        - native_events(bool)(False)
           |  Whether to expect incoming events to be native Wishbone events.

        - initial_listing(bool)(True)

           |  When True, generates for each defined path an event.  This is
           |  useful to initially give depending modules the filenames they
           |  need to function.

        - glob_pattern(str)(*)

           |  A glob pattern to filter out only matching files.

        - paths(dict)({"/tmp": ["IN_CREATE", "IN_CLOSE_WRITE"]})

           |  A dict of paths with a list of inotify events to monitor.  When
           |  the list is empty no filtering is done and results into all
           |  inotify events going through.


    Queues::

        - outbox
           |  Outgoing notify events.

    '''

    def __init__(self, actor_config, native_events=False, destination="data",
                 initial_listing=True, glob_pattern="*", paths={"/tmp": ["IN_CREATE", "IN_CLOSE_WRITE"]}):
        InputModule.__init__(self, actor_config)

        self.pool.createQueue("outbox")

    def preHook(self):

        for path, event_types in self.kwargs.paths.items():
            for event_type in event_types:
                if event_type not in constants.__dict__:
                    raise Exception("Inotify event type '%s' defined for path '%s' is not valid." % (event_type, path))

        for path, inotify_types in self.kwargs.paths.items():
            self.sendToBackground(self.monitor, path, inotify_types, self.kwargs.glob_pattern)

    def monitor(self, path, inotify_types, glob_pattern):
        '''Monitors ``path`` for ``inotify_types`` on files and dirs matching ``glob_pattern``

        :param str path: The path to monitor
        :param list inotify_types: A list of inotify types to monitor
        :param str glob_pattern: Paths need to match in order to be returned.
        :return: None
        '''

        all_types = ', '.join(inotify_types)
        if all_types == '':
            all_types = "ALL"

        while self.loop():
            if os.path.exists(path) and os.access(path, os.R_OK):
                self.logging.info("Started to monitor path '%s' for '%s' inotify events on paths matching '%s'." % (os.path.abspath(path), all_types, glob_pattern))

                if self.kwargs.initial_listing:
                    for p in self.__getAllFiles(path, glob_pattern):
                        for payload in self.decode(p):
                            event = self.generateEvent({"path": os.path.abspath(payload), "inotify_type": "WISHBONE_INIT"}, self.kwargs.destination)
                            self.submit(event, "outbox")
                try:
                    for abs_path, i_type in self.__setupInotifyMonitor(path, inotify_types, glob_pattern):
                        for payload in self.decode(abs_path):
                            event = self.generateEvent({"path": payload, "inotify_type": i_type}, self.kwargs.destination)
                            self.submit(event, "outbox")
                except Exception as err:
                    self.logging.critical('Failed to initialize inotify monitor. This needs immediate attention. Reason: %s' % err)
                    sleep(1)
            else:
                self.logging.warning("The defined path '%s' does not exist or is not readable. Will sleep for 5 seconds and try again." % (path))
                sleep(5)

    def __getAllFiles(self, path, glob_pattern):

        """
        Returns all files it can find in ``path``.  If ``path`` is a directory
        it returns all files found in the directory (not recursive)

        :param str path: The path to investigate.
        :param str glob_pattern: The patterns to which files and dirs have to match.
        :return: Generator
        """

        if os.path.isdir(path):
            all_files = [os.path.abspath("%s/%s" % (path, name)) for name in os.listdir(path) if os.path.isfile("%s/%s" % (path, name))]
        else:
            all_files = [os.path.abspath(path)]

        for f in all_files:
            if fnmatch.fnmatch(f, glob_pattern):
                yield f

    def __setupInotifyMonitor(self, path, inotify_types, glob_pattern):

        '''

        Initializes an inotify monitor process on ``path`` and yields the
        defined ``inotify_types`` generated on the paths matching the
        ``glob_pattern``.

        :param str path: The path to monitor
        :param list inotify_types: A list of inotify types to monitor
        :param str glob_pattern: Paths need to match in order to be returned.
        :return: generator

        '''

        file_exists = True
        i = Inotify()
        i.add_watch(path)

        while file_exists and self.loop():
            for event in i.event_gen(yield_nones=False):
                if event is not None:
                    for inotify_type in event[1]:
                        if inotify_type in inotify_types or inotify_types == []:
                            abs_path = os.path.abspath("%s/%s" % (event[2], event[3]))
                            if fnmatch.fnmatch(abs_path, glob_pattern):
                                yield abs_path.rstrip('/'), inotify_type
                        if inotify_type == "IN_DELETE_SELF":
                            file_exists = False
                            break
                else:
                    sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  modify.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.actor import Actor
from wishbone.module import ProcessModule
from copy import deepcopy
import re
import arrow

VALID_EXPRESSIONS = ["add_item",
                     "copy",
                     "del_item",
                     "delete",
                     "extract",
                     "join",
                     "lowercase",
                     "merge",
                     "replace",
                     "set",
                     "template",
                     "uppercase",
                     "time",
                     ]


class Modify(ProcessModule):

    '''**Modify and manipulate datastructures.**

    This module modifies the data of an event using a **sequential** list of
    expressions.

    Expressions are dictionaries containing 1 item. The key is a string and
    the value is a list of parameters accepted by the expression.

    For example::

        {"set": ["hi", "data.one"]}


    Sets the value *"hi"* to key *"data.one"*.

    In the the YAML formatted bootstrap file that would look like::

        module: wishbone.function.modify
        arguments:
          expressions:
            - set: [ "hi", "data.one" ]


    Valid expressions are:


      - **add_item**::

          add_item: [<item>, <key>]

        Adds *<item>* to the list stored under *<key>*.


      - **copy**::

          copy: [<source_key>, <destination_key>, <default_value>]

        Copies *<source_key>* to *<destination_key>* and overwrites
        *<destination_key>* when it exists.  If <source_key> does not exist,
        <default_value> is taken instead.


      - **del_item**::

          del_item: [<key>, <item>]

        Deletes first occurance of *<item>* from the list stored under
        *<source_key>*.


      - **delete**::

          delete: [<key>]

        Deletes *<key>* from the event.


      - **extract**::

          extract: [<destination>, <regex>, <source>]

        Makes use of Python re module to extract named groups from *<source>*
        using *<regex>* and add the resulting matches to *<destination>*.

        The following example would extract the words "one" and "two" from
        "data.test" and add the to data.extract:

          expression::

            extract: ["data.extract", '(?P<first>.*?);(?P<second>.*)', "data.test"]

          result::

            {"data":{"test:"one;two", extract:{"first": "one", "second": "two"}}}

      - **join**::

          join: [<array>, <join>, <destination>]

        Joins an array into a string using the <join> value.


      - **lowercase**::

          lowercase: [<key>]

        Turns the string stored under *<key>* to lowercase.

      - **merge**::

          merge: [<object_one>, <object_two>, <destination>]

        Merges 2 arrays into <destination>


      - **replace**::

          replace: [<regex>, <value>, <key>]

        replaces every occurance of <regex> of the value stored in <key> with <value>


      - **set**::

          set: [<value>, <key>]

        Sets *<value>* to the event *<key>*.


      - **uppercase**::

          uppercase: [<key>]

        Turns the string stored under *<key>* to uppercase.


      - **template**::

          template: [<destination_key>, <template>, <source_key>]

        Uses the dictionary stored in *<source_key>* to complete *<template>*
        and stores the results into key *<destination_key>*. The templating
        language used is Python's builtin string format one.


      - **time**::

          time: [<destination_key>, <format>]

        Modifies the *<timestamp>* value according the the *<format>* specification
        and stores it into *<destination_key>*.
        See http://crsmithdev.com/arrow/#format for the format.



    Parameters:

        - expressions(list)([])
           |  A list of expressions to apply.

    Queues:

        - inbox:
           |  Incoming messages

        - outbox:
           |  Outgoing modified messages
    '''

    def __init__(self, actor_config, expressions=[]):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.registerConsumer(self.consume, "inbox")

    def consume(self, event):

        for expression in self.kwargs.expressions:
            try:
                command, args = self.__extractExpr(expression)
                event = getattr(self, "command_%s" % (command))(event, *args)
            except Exception as err:
                self.logging.error("Failed to process expression '%s'. Reason: %s Skipped." % (expression, err))

        self.submit(event, "outbox")

    def command_add_item(self, event, item, key):

        event.get(key).append(item)
        return event

    def command_copy(self, event, source, destination, default_value):

        try:
            event.copy(source, destination)
        except KeyError:
            event.set(default_value, destination)
        return event

    def command_del_item(self, event, item, key):

        event.get(key).remove(item)

        return event

    def command_delete(self, event, key):

        event.delete(key)
        return event

    def command_extract(self, event, destination, regex, source):

        matches = re.match(regex, event.get(source))
        event.set(matches.groupdict(), destination)
        return event

    def command_join(self, event, array, j, destination):

        result = j.join(event.get(array))
        event.set(result, destination)
        return event

    def command_lowercase(self, event, key):

        event.set(event.get(key).lower(), key)
        return event

    def command_merge(self, event, one, two, destination):
        result = event.get(one) + event.get(two)
        event.set(result, destination)
        return event

    def command_replace(self, event, regex, value, key):

        result = re.sub("{}".format(regex), value, str(event.get(key)))
        event.set(str(result), key)
        return event

    def command_set(self, event, value, key):

        event.set(deepcopy(value), key)
        return event

    def command_template(self, event, destination, template, key):

        result = template.format(**event.get(key))
        event.set(result, destination)
        return event

    def command_time(self, event, destination_key, f):

        result = arrow.get(event.get("timestamp")).format(f)
        event.set(result, destination_key)
        return event

    def command_uppercase(self, event, key):

        event.set(event.get(key).upper(), key)
        return event

    def __extractExpr(self, e):

        assert isinstance(e, dict), "The expression should be a dict."
        assert len(list(e.keys())) == 1, "The expression should only contain 1 value"

        c = list(e.keys())[0]
        assert c in VALID_EXPRESSIONS, "'%s' is an invalid expression." % c

        assert isinstance(e[c], list), "The expression value must be a list."
        return c, e[c]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  default.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.actorconfig import ActorConfig
from wishbone.error import ModuleInitFailure, NoSuchModule
from wishbone.error import QueueConnected
from wishbone.componentmanager import ComponentManager
from gevent import event, sleep, spawn
from gevent import pywsgi
from .graphcontent import GRAPHCONTENT
from .graphcontent import VisJSData
from types import SimpleNamespace


class ModulePool():

    def __init__(self):

        self.module = SimpleNamespace()

    def list(self):
        '''Returns a generator returning all module instances.'''

        for m in list(self.module.__dict__.keys()):
            yield self.module.__dict__[m]

    def getModule(self, name):
        '''Returns a module instance'''

        try:
            return getattr(self.module, name)
        except AttributeError:
            raise NoSuchModule("Could not find module %s" % name)

    def hasModule(self, name):
        '''
        Checks whether the module pool has this module.


        Args:
            name (str): The name of the module instance.

        Returns
            bool: True if module exists False if not.
        '''

        return name in self.module.__dict__.keys()


class Default(object):

    '''
    The default Wishbone router.

    A Wishbone router is responsible for organising the event flow between
    modules.

    Args:
        config (EasyDict): The router setup configuration.
        size (int): The size of all queues.
        frequency (int)(1): The frequency at which metrics are produced.
        identification (wishbone): A string identifying this instance in logging.
    '''

    def __init__(self, config=None, size=100, frequency=10, identification="wishbone", graph=False, graph_include_sys=False):

        self.component_manager = ComponentManager()
        self.config = config
        self.size = size
        self.frequency = frequency
        self.identification = identification
        self.graph = graph
        self.graph_include_sys = graph_include_sys

        self.module_pool = ModulePool()
        self.__block = event.Event()
        self.__block.clear()

        self.__connections = {
        }

    def block(self):
        '''Blocks until stop() is called and the shutdown process ended.'''

        self.__block.wait()

    def connectQueue(self, source, destination):
        '''Connects one queue to the other.

        For convenience, the syntax of the queues is <modulename>.<queuename>
        For example:

            stdout.inbox

        This type of router actually replaces the destination queue with the source
        queue.

        Args:
            source (str): The source queue in <module.queue_name> syntax
            destination (str): The destination queue in <module.queue_name> syntax
        '''

        (source_module, source_queue) = source.split('.')
        (destination_module, destination_queue) = destination.split('.')

        if not self.module_pool.hasModule(source_module):
            raise NoSuchModule("Module instance %s does not exist." % (source_module))

        if not self.module_pool.hasModule(destination_module):
            raise NoSuchModule("Module instance %s does not exist." % (destination_module))

        result = self.__isConnectedTo(source)
        if result is not None:
            raise QueueConnected("Queue %s is already connected to %s." % (source, result))

        result = self.__isConnectedTo(destination)
        if result is not None:
            raise QueueConnected("Queue %s is already connected to %s." % (destination, result))

        self.__connections[source] = destination

        source_module_instance = self.module_pool.getModule(source_module)
        if not source_module_instance.pool.hasQueue(source_queue):
            source_module_instance.pool.createSystemQueue(source_queue)
            source_module_instance.logging.debug("Module instance '%s' has no queue '%s' so auto created." % (source_module, source_queue))

        destination_module_instance = self.module_pool.getModule(destination_module)
        if not destination_module_instance.pool.hasQueue(destination_queue):
            destination_module_instance.pool.createSystemQueue(destination_queue)
            destination_module_instance.logging.debug("Module instance '%s' has no queue '%s' so auto created." % (destination_module, destination_queue))

        setattr(
            destination_module_instance.pool.queue,
            destination_queue,
            source_module_instance.pool.getQueue(
                source_queue
            )
        )

        source_module_instance.pool.getQueue(source_queue).disableFallThrough()
        source_module_instance.logging.debug("Connected queue %s to %s" % (source, destination))

    def getChildren(self, module):
        '''
        Returns all the connected child modules

        Args:
            module (str): The name of the module.

        Returns:
            list: A list of module names.
        '''

        children = []

        def travel(m):
            for connection in self.__connections:
                if connection.split('.')[0] == m:
                    child = self.__connections[connection].split('.')[0]
                    if child in children:
                        continue
                    else:
                        children.append(child)
                        travel(child)

        travel(module)
        return children

    def registerModule(self, module, actor_config, arguments={}):
        '''Initializes the wishbone module ``module``.

        Args:
            module (str): A Wishbone module component name.
            actor_config (ActorConfig): The module's actor configuration
            arguments (dict): The parameters to initialize the module.
        '''

        try:
            m = self.component_manager.getComponentByName(module)
            setattr(self.module_pool.module, actor_config.name, m(actor_config, **arguments))
        except Exception as err:
            raise ModuleInitFailure("Problem loading module '%s'.  Reason: %s" % (actor_config.name, err))

    def stop(self):
        '''Stops all running modules.'''

        for module in self.module_pool.list():
            if module.name not in list(self.getChildren("_logs")) + ["_logs"] and not module.stopped:
                module.stop()

        while not self.__logsEmpty():
            sleep(0.1)

        self.__running = False
        self.__block.set()

    def start(self):
        '''Starts all registered modules.'''

        if self.config is not None:
            self.__initConfig()

        if self.graph:
            self.graph = GraphWebserver(self.config, self.module_pool, self.__block, self.graph_include_sys)
            self.graph.start()

        for module in self.module_pool.list():
            module.start()

    def __initConfig(self):
        '''Setup all modules and routes.'''

        protocols = {}
        for name, instance in list(self.config.protocols.items()):
            protocols[name] = self.component_manager.getComponentByName(instance.protocol)(**instance.arguments)

        template_functions = {}
        for name, instance in list(self.config.template_functions.items()):
            template_functions[name] = self.component_manager.getComponentByName(instance.function)(**instance.arguments)

        module_functions = {}
        for name, instance in list(self.config.module_functions.items()):
            module_functions[name] = self.component_manager.getComponentByName(instance.function)(**instance.arguments)

        for name, instance in list(self.config.modules.items()):
            mod_func = {}
            for queue, queue_functions in list(instance.functions.items()):
                mod_func[queue] = []
                for queue_function in queue_functions:
                    if queue_function in module_functions:
                        mod_func[queue].append(module_functions[queue_function])

            if instance.protocol is None:
                protocol = None
            elif instance.protocol not in protocols:
                raise ModuleInitFailure("Protocol %s referenced but not available." % (instance.protocol))
            else:
                protocol = protocols[instance.protocol]

            actor_config = ActorConfig(
                name=name,
                size=self.size,
                frequency=self.frequency,
                template_functions=template_functions,
                description=instance.description,
                module_functions=mod_func,
                identification=self.identification,
                protocol=protocol,
                io_event=instance.event
            )

            self.registerModule(
                instance.module,
                actor_config,
                instance.arguments
            )

        self.__setupConnections()

    def __isConnectedTo(self, queue):
        '''
        Returns the module.queue ``queue`` is connected to.

        Args:
            queue (str): The name of the queue in ``module.queue`` format.

        Returns
            str/None: The name of the queue which is connected.
        '''

        inverse = {v: k for k, v in self.__connections.items()}

        if queue in self.__connections:
            return self.__connections[queue]
        elif queue in inverse:
            return inverse[queue]
        else:
            return None

    def __logsEmpty(self):
        '''Checks each module whether any logs have stayed behind.'''

        for module in self.module_pool.list():
            if not module.pool.queue._logs.size() == 0:
                return False
        else:
            return True

    def __setupConnections(self):
        '''Setup all connections as defined by configuration_manager'''

        for route in self.config.routingtable:
            self.connectQueue("%s.%s" % (route.source_module, route.source_queue), "%s.%s" % (route.destination_module, route.destination_queue))


class GraphWebserver():

    def __init__(self, config, module_pool, block, include_sys):
        self.config = config
        self.module_pool = module_pool
        self.block = block
        self.include_sys = include_sys
        self.js_data = VisJSData()

        for c in self.config["routingtable"]:
            if not self.include_sys and any([
                    c.source_module.startswith('_'),
                    c.destination_module.startswith('_'),
                    c.source_queue.startswith('_'),
                    c.destination_queue.startswith('_')]):
                continue
            else:
                self.js_data.addModule(instance_name=c.source_module,
                                       module_name=self.config["modules"][c.source_module]["module"],
                                       description=self.module_pool.getModule(c.source_module).description)

                self.js_data.addModule(instance_name=c.destination_module,
                                       module_name=self.config["modules"][c.destination_module]["module"],
                                       description=self.module_pool.getModule(c.destination_module).description)

                self.js_data.addQueue(c.source_module, c.source_queue)
                self.js_data.addQueue(c.destination_module, c.destination_queue)
                self.js_data.addEdge("%s.%s" % (c.source_module, c.source_queue), "%s.%s" % (c.destination_module, c.destination_queue))

    def start(self):

        print("#####################################################")
        print("#                                                   #")
        print("# Caution: Started webserver on port 8088           #")
        print("#                                                   #")
        print("#####################################################")
        spawn(self.setupWebserver)

    def stop(self):
        pass

    def loop(self):

        return self.__block

    def application(self, env, start_response):
        if env['PATH_INFO'] == '/':
            start_response('200 OK', [('Content-Type', 'text/html')])
            return[str.encode(GRAPHCONTENT % (self.js_data.dumpString()[0], self.js_data.dumpString()[1]))]
        else:
            start_response('404 Not Found', [('Content-Type', 'text/html')])
            return [b'<h1>Not Found</h1>']

    def setupWebserver(self):

        pywsgi.WSGIServer(('', 8088), self.application, log=None, error_log=None).serve_forever()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  structured_data_file.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


import os
import json
import yaml
from gevent.lock import Semaphore
from jsonschema import validate
from wishbone.error import InvalidData


class StructuredDataFile():
    '''
    Loads and optionally validates structured data from disk.

    The module loads data from disk and optionally validates the data against
    the provided schema.  The datastructure can be JSON and/or YAML.

    Attributes:
        default (obj): The default value when no file content has loaded yet.
        schema (str): The JSONschema to validate the loaded data against.
        expect_json (bool): When True, the data structure can be JSON
        expect_yaml (bool): When True, the data structure can be YAML.
        expect_kv (bool): When True, the data structure can be Key/Value.
        content (dict): The loaded configurations.  Absolute paths are dict keys.
    '''

    def __init__(self, default=None, schema=None, expect_json=True, expect_yaml=True, expect_kv=True):
        '''
        Args:
            default (obj): A default value to return when no file content has loaded yet.
            schema (str): The JSONschema to validate the loaded data against.
            expect_json (bool): When True, the data structure can be JSON
            expect_yaml (bool): When True, the data structure can be YAML.
            expect_kv (bool): When True, the data structure can be Key/Value.
        '''

        self.default = default
        self.schema = schema
        self.expect_json = expect_json
        self.expect_yaml = expect_yaml
        self.expect_kv = expect_kv
        self.content = {}
        self.lock = Semaphore()

    def delete(self, path):
        '''Deletes the file content from the object'''

        abs_path = os.path.abspath(path)

        with self.lock:
            if abs_path in self.content.keys():
                del(self.content[abs_path])

    def dump(self):
        '''Dumps the complete cached content of all files
        If multiple files are cached then a list is returned.
        '''

        with self.lock:
            if self.content is {}:
                return self.default
            else:
                return self.content

    def dumpItems(self):
        '''
        Dumps the complete cached content file by file
        '''

        for filename, content in self.content.items():
            yield content

    def get(self, path):
        '''Returns the cached content of the file.  If the file isn't loaded yet, it
        tries to do that.'''

        abs_path = os.path.abspath(path)

        with self.lock:
            if abs_path not in self.content.keys():
                self.content[abs_path] = self.__load(abs_path)
            return self.content[abs_path]

    def load(self, path):
        '''Loads the file into the module and validates the content when required.'''

        abs_path = os.path.abspath(path)
        with self.lock:
            self.content[abs_path] = self.__load(abs_path)

    def __load(self, abs_path):

        if os.path.exists(abs_path) and os.access(abs_path, os.R_OK):
            if os.path.isfile(abs_path):
                content = self.__readFile(abs_path)
                if self.schema is not None:
                    try:
                        validate(content, self.schema)
                    except Exception as err:
                        raise InvalidData(err.message)
            else:
                raise Exception("'%s' does not appear to be a regular file." % (abs_path))
        else:
            raise Exception("File '%s' does not exist or is not accessible." % (abs_path))

        return content

    def __readFile(self, path):

        with open(path) as f:
            errors = []

            if self.expect_json:
                try:
                    f.seek(0)
                    return json.load(f)
                except Exception as err:
                    errors.append("JSON: %s" % str(err))
                else:
                    return

            if self.expect_yaml:
                try:
                    f.seek(0)
                    return yaml.load(f)
                except Exception as err:
                    errors.append("YAML: %s" % str(err))
                else:
                    return

            if self.expect_kv:
                try:
                    data = {}
                    f.seek(0)
                    for line in f.readlines():
                        splitted = line.split(":")
                        key = splitted[0]
                        value = ":".join(splitted[1:]).rstrip().lstrip()
                        data[key] = value
                    return data
                except Exception as err:
                    errors.append("KV: %s" % str(err))
                else:
                    return

            if len(errors) > 0:
                raise Exception("Could not load file '%s'.  Reason: '%s'" % (path, ",".join(errors)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  json.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.protocol import Encode
from json import dumps


class JSON(Encode):

    '''**Encode data into JSON format.**

    Convert a Python datastructure into JSON format.

    Parameters:

        - sort_keys(bool)(False)
            |  Sorts keys when True

        - ensure_ascii(bool)(False)
            |  When True makes sure all chars are valid ascii

        - ident(int)(None)
            |  The indentation used.

    '''

    def __init__(self, sort_keys=False, ensure_ascii=False, indent=None):
        self.sort_keys = sort_keys
        self.ensure_ascii = ensure_ascii
        self.indent = indent

    def handleDict(self, data):
        return dumps(data, sort_keys=self.sort_keys, ensure_ascii=self.ensure_ascii, indent=self.indent)

    def handleCut(self, data):
        return self.handleDict(dict(data))

    handleList = handleDict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  set.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone.function.module import ModuleFunction


class Set(ModuleFunction):
    '''
    Sets a field to the desired value.

    A Wishbone module function which sets data to the desired field.  Data can be a template.

    Args:
        data (str): The value (or template) to apply.
        destination (str): The destination field
    '''

    def __init__(self, data, destination='data'):

        self.data = data
        self.destination = destination

    def do(self, event):
        '''
        The function mapped to the module function.

        Args:
            event (wishbone.event.Event): The Wishbone event.

        Returns:
            wishbone.event.Event: The modified event.
        '''

        if isinstance(self.data, str):
            data = event.render(self.data)
        else:
            data = self.data

        event.set(data, self.destination)

        return event



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  bootstrap.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import monkey; monkey.patch_all()
import gipc
import argparse
import os
import sys
# http://stackoverflow.com/questions/4554271/how-to-avoid-excessive-stat-etc-localtime-calls-in-strftime-on-linux
os.environ["TZ"] = ":/etc/localtime"

from wishbone.router.default import Default
from wishbone.componentmanager import ComponentManager
from wishbone.config import ConfigFile
from wishbone.utils import PIDFile
from gevent import signal
from gevent.event import Event
from daemon import DaemonContext
from pkg_resources import get_distribution


class BootStrap():
    '''
    Parses command line arguments and bootstraps the Wishbone instance.
    '''

    def __init__(self, description="Pragmatic event processing servers.", include_groups=[]):

        parser = argparse.ArgumentParser(description=description)
        subparsers = parser.add_subparsers(dest='command')
        subparsers.required = True

        start = subparsers.add_parser('start', description="Bootsraps a Wishbone instance.")
        start.add_argument('--config', type=str, dest='config', default='wishbone.cfg', help='The Wishbone bootstrap file to load.')
        start.add_argument('--frequency', type=int, dest='frequency', default=10, help='The metric frequency.')
        start.add_argument('--graph', action="store_true", dest="graph", help='When enabled starts a webserver on 8088 showing a graph of connected modules and queues.')
        start.add_argument('--graph-include-sys', action="store_true", dest="graph_include_sys", help='When enabled includes logs and metrics related queues modules and queues to graph layout.')
        start.add_argument('--identification', type=str, dest='identification', default="wishbone", help='An identifier string for generated logs.')
        start.add_argument('--instances', type=int, dest='instances', default=1, help='The number of parallel Wishbone instances to bootstrap.')
        start.add_argument('--loglevel', type=int, dest='log_level', default=6, help='The maximum loglevel.')
        start.add_argument('--fork', action="store_true", default=False, help="When defined forks Wishbone to background and INFO logs are written to syslog/journald.")
        start.add_argument('--nocolor', action="store_true", help='When defined does not print colored output to stdout.')
        start.add_argument('--pid', type=str, dest='pid', default='%s/wishbone.pid' % (os.getcwd()), help='The pidfile to use.')
        start.add_argument('--profile', action="store_true", help='When enabled profiles the process and dumps a Chrome developer tools profile file in the current directory.')
        start.add_argument('--queue-size', type=int, dest='queue_size', default=100, help='The queue size to use.')

        stop = subparsers.add_parser('stop', description="Tries to gracefully stop the Wishbone instance.")
        stop.add_argument('--pid', type=str, dest='pid', default='wishbone.pid', help='The pidfile to use.')

        list = subparsers.add_parser('list', description="Lists the available modules.")
        list.add_argument('--namespace', type=str, dest='namespace', default="wishbone, wishbone_contrib, wishbone_external", help='The component namespace to query.')

        show = subparsers.add_parser('show', description="Shows information about a component.")
        show_group = show.add_mutually_exclusive_group(required=True)
        show_group.add_argument('--docs', type=str, help='Shows the documentation of the component.')
        show_group.add_argument('--code', type=str, help='Shows the code of the refered component.')

        arguments = vars(parser.parse_args())

        dispatch = Dispatch(**arguments)
        getattr(dispatch, arguments["command"])()


class Dispatch():
    '''
    Handles the Wishbone commands, processes and daemons.
    '''

    def __init__(self, **kwargs):
        self.command = kwargs.get("command", None)
        self.config = kwargs.get("config", None)
        self.instances = kwargs.get("instances", None)
        self.pid = kwargs.get("pid", None)
        self.queue_size = kwargs.get("queue_size", None)
        self.frequency = kwargs.get("frequency", None)
        self.identification = kwargs.get("identification", None)
        self.graph = kwargs.get("graph", None)
        self.graph_include_sys = kwargs.get("graph_include_sys", None)
        self.profile = kwargs.get("profile", None)
        self.docs = kwargs.get("docs", None)
        self.code = kwargs.get("code", None)
        self.nocolor = kwargs.get("nocolor", False)
        self.fork = kwargs.get("fork", None)
        self.log_level = kwargs.get("log_level", None)
        self.namespace = kwargs.get("namespace", None)
        self.routers = []

    def bootstrapBlock(self):
        '''Helper function which blocks untill all running routers have stopped.
        '''
        while True:
            try:
                for router in self.routers:
                    router.join()
                break
            except KeyboardInterrupt:
                pass

    def generateHeader(self):
        '''Generates the Wishbone ascii header.
        '''

        with open("%s/data/banner.tmpl" % (os.path.dirname(__file__))) as f:
            template = ''.join(f.readlines()).format(version=get_distribution('wishbone').version)

        return template

    def initializeOneRouter(self, config):
        '''Initializes a Router instance using the provided config object.

        This function blocks until signal(2) is received after which it
        continues and executes the router's stop() method.

        The idea is that you can spawn this function to the background to have
        multiple parallel instances.

        Args:
            config (Wishbone.config.configfile:ConfigFile): The router configuration
        '''

        def startRouter():
            router = Default(
                config,
                size=self.queue_size,
                frequency=self.frequency,
                identification=self.identification,
                graph=self.graph,
                graph_include_sys=self.graph_include_sys
            )

            router.start()

            e.wait()

            router.stop()
        e = Event()
        signal(2, e.set)

        if self.profile:
            from wishbone.utils.py2devtools import Profiler
            with Profiler():
                startRouter()
        else:
            startRouter()

    def initializeManyRouters(self, config, number, background):

        '''Initialize many routers and background if required

        Args:
            config (Wishbone.config.configfile:ConfigFile): The router configration
            number (int): The number of instances to intialize
            background (bool): Whether to background the routers or not
        '''

        if background:
            pid_file = PIDFile(self.pid)
            with DaemonContext(stdout=sys.stdout, stderr=sys.stderr, detach_process=True):
                if self.instances == 1:
                    sys.stdout.write("\nWishbone instance started with pid %s\n" % (os.getpid()))
                    sys.stdout.flush()
                    pid_file.create([os.getpid()])
                    self.initializeOneRouter(config)
                else:
                    for instance in range(self.instances):
                        self.routers.append(
                            gipc.start_process(
                                self.initializeOneRouter,
                                args=(config, ),
                                daemon=True
                            )
                        )

                    pids = [str(p.pid) for p in self.routers]
                    print(("\nInstances started in foreground with pid %s\n" % (", ".join(pids))))
                    pid_file.create(pids)

                self.bootstrapBlock()
        else:
            if self.instances == 1:
                sys.stdout.write("\nInstance started in foreground with pid %s\n" % (os.getpid()))
                self.initializeOneRouter(config)
            else:
                for instance in range(self.instances):
                    self.routers.append(
                        gipc.start_process(
                            self.initializeOneRouter,
                            args=(config, ),
                            daemon=True
                        )
                    )

                pids = [str(p.pid) for p in self.routers]
                print(("\nInstances started in foreground with pid %s\n" % (", ".join(pids))))
                self.bootstrapBlock()

    def list(self):
        '''Maps to the CLI command and lists all Wishbone entrypoint modules it can find.
        '''

        print((self.generateHeader()))
        print((ComponentManager(namespace=[n.strip() for n in self.namespace.split(",")]).getComponentTable()))

    def show(self):
        '''Maps to the CLI command and shows the docstring of the Wishbone module.
        '''

        component_manager = ComponentManager()

        if self.docs is not None:
            component_manager.validateComponentName(self.docs)
            component_manager.exists(self.docs)

            print((self.generateHeader()))
            (namespace, component_type, category, name) = self.docs.split('.')

            try:
                title = component_manager.getComponentTitle(namespace, component_type, category, name)
                version = component_manager.getComponentVersion(namespace, component_type, category, name)
                header = "%s.%s.%s.%s" % (namespace, component_type, category, name)
                print("")
                print(("=" * len(header)))
                print(header)
                print(("=" * len(header)))
                print("")
                print(("Version: %s" % (version)))
                print("")
                print(title)
                print(("-" * len(title)))
                print((component_manager.getComponentDoc(namespace, component_type, category, name)))
            except Exception as err:
                print("Failed to load component '%s'. Reason: %s" % (self.docs, err))

        if self.code is not None:

            component_manager.validateComponentName(self.code)
            component_manager.exists(self.code)

            import inspect
            from pygments import highlight
            from pygments.lexers import get_lexer_by_name
            from pygments.formatters import terminal

            component = component_manager.getComponentByName(self.code)
            code = "".join(inspect.getsourcelines(component)[0])
            print(highlight(code, get_lexer_by_name("python"), terminal.TerminalFormatter()))

    def start(self):
        '''Maps to the CLI command and starts one or more Wishbone processes in background.
        '''

        colorize = not self.nocolor

        if self.fork:
            logstyle = "SYSLOG"
        else:
            logstyle = "STDOUT"

        router_config = ConfigFile(
            filename=self.config,
            logstyle=logstyle,
            loglevel=self.log_level,
            colorize_stdout=colorize,
            identification=self.identification
        )
        config = router_config.dump()

        self.initializeManyRouters(
            config=config,
            number=self.instances,
            background=self.fork
        )

    def stop(self):
        '''Maps to the CLI command and stop the running Wishbone processes.
        '''

        try:
            pid = PIDFile(self.pid)
            sys.stdout.write("Stopping instance with PID ")
            sys.stdout.flush()
            for entry in pid.read():
                sys.stdout.write(" %s " % (entry))
                sys.stdout.flush()
                pid.sendSigint(entry)
            pid.cleanup()
            print("")
        except Exception as err:
            print("")
            print(("Failed to stop instances.  Reason: %s" % (err)))


def main():
    try:
        BootStrap()
    except Exception as err:
        print(("Failed to bootstrap instance.  Reason: %s" % (err)))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  actor.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent.monkey import patch_all; patch_all()
from wishbone.queue import QueuePool
from wishbone.logging import Logging
from wishbone.event import Event as Wishbone_Event
from wishbone.error import ModuleInitFailure, InvalidModule, TTLExpired
from wishbone.actorconfig import ActorConfig
from wishbone.function.template import TemplateFunction
from wishbone.function.module import ModuleFunction

from collections import namedtuple
from gevent import spawn, kill
from gevent import sleep, socket
from gevent.event import Event
from wishbone.error import QueueFull
from time import time
from sys import exc_info
import traceback
import inspect
import jinja2
from copy import deepcopy

from easydict import EasyDict
from pkg_resources import get_distribution


Greenlets = namedtuple('Greenlets', "consumer generic log metric")


class Actor(object):
    """A base class providing core Actor functionality.

    The Actor base class is responsible for providing the base functionality,
    setup and helper functions of a Wishbone module.

    Args:
        config (wishbone.actorconfig.ActorConfig): The ActorConfig object instance.

    Attributes:
        config (wishbone.actorconfig.ActorConfig): The ActorConfig object instance.
        name (str): The name of the instance, derived from `config`.
        description (str): The description of the actor based instance, derived from `config`.

                pool (wishbone.pool.QueuePool): The Actor's queue pool.
    Methods:
        logging (wishbone.logging.Logging)

    """

    def __init__(self, config):

        self.config = config
        self.name = config.name
        self.description = self.__getDescription(config)

        self.pool = QueuePool(config.size)

        self.logging = Logging(
            name=config.name,
            q=self.pool.queue._logs,
            identification=self.config.identification
        )

        self.__loop = True
        self.greenlets = Greenlets([], [], [], [])
        self.greenlets.metric.append(spawn(self.__metricProducer))

        self._run = Event()
        self._run.clear()

        self.stopped = True

        # Setup the Jinja2 environment to render kwargs templates.
        ##########################################################
        self.env_template = jinja2.Environment(
            undefined=jinja2.StrictUndefined,
            trim_blocks=True,
            loader=jinja2.FileSystemLoader('/')
        )

        # Add the template functions to the template globals
        ####################################################
        for key, value in self.config.template_functions.items():
            self.env_template.globals.update({key: value.get})

        # Store a copy of the raw/unmodified kwargs
        ##########################################
        self.kwargs_raw = self.__getRawKwargs()

        # Store a copy of kwargs with all templates replaced by a template instance
        ############################################################################
        self.kwargs_template = self.__getTemplateKwargs(self.env_template, self.kwargs_raw)

        # Store a copy of the rendered kwargs as an EasyDict instance
        #############################################################
        self.kwargs = self.__renderTemplateKwargs(self.kwargs_template)

        # Do some sanity checks
        #######################
        self.__sanityChecks()

        # Validate and setup module
        # Methods are come from based class
        ###################################
        self._moduleInitValidation()
        self._moduleInitSetup()

        self.version = self.__getVersion()

    def generateEvent(self, data={}, destination=None):
        '''
        Generates a new event.

        This function can get overridden by
        ``wishbone.module.InputModule._generateNativeEvent``.

        The provided ``data`` will be traversed in search of valid templates
        which then will be rendered.

        Args:
            data (``data``): The payload to add to the event.
            destination (None): The destination key to write the data to

        Returns:
            wishbone.event.Event: An event containing ``data`` as a payload.

        '''
        if destination in [None, "data"]:
            event = Wishbone_Event(data)
            event.renderField(destination, self.env_template)
        else:
            event = Wishbone_Event()
            event.set(data, destination)
            event.renderField(destination, self.env_template)
        return event

    def loop(self):
        '''The global lock for this module.

        Returns:
            bool: True when module is in running mode. False if not.
        '''

        return self.__loop

    def postHook(self):
        '''
        Is executed when module exits.
        '''

        self.logging.debug("Module has no postHook() method set.")

    def preHook(self):
        '''
        Is executed when module starts. Can be overriden by the user.
        '''

        self.logging.debug("Module has no preHook() method set.")

    def registerConsumer(self, function, queue):
        '''
        Registers <function> to process all events in <queue>

        Don't not trap errors here.  When <function> fails then the event will be
        submitted to the "failed" queue,  If <function> succeeds to the
        success queue.

        Registering ``function`` to consume ``queue`` will also apply all the
        registered module functions against the events consumed from it.

        Args:
            function (``function``): The function which processes events
            queue (str): The name of the queue from which ``function`` will
                         process the events.

        Returns:
            None
        '''

        self.greenlets.consumer.append(spawn(self._consumer, function, queue))

    def renderEventKwargs(self, event, queue=None):
        '''
        Renders kwargs using the content of ``event`` and stores the result under
        ``event.kwargs``.

        Args:
            event (``wishbone.event.Event``): An Event instance
            queue (str): The queue name so ``RenderKwargs`` can store the results
                         in the correct queue context.

        Returns:
            ``wishbone.event.Event``: The provided event instance.
        '''

        event.kwargs = self.__renderKwargs.render(
            queue_context=queue,
            event_content=event.dump()
        )
        return event

    def renderKwargs(self):
        '''
        Renders kwargs without making use of event content. This is typically
        used when initiliazing a module and render the defined kwargs which do
        not need a event data for rendering.

        Returns:
            None
        '''

        self.kwargs = self.__renderKwargs.render()

    def start(self):
        '''
        Starts the module.

        Returns:
            None
        '''

        self.__postHook()
        if hasattr(self, "preHook"):
            self.logging.debug("preHook() found, executing")
            self.preHook()
        self.__validateAppliedFunctions()
        self._run.set()
        self.logging.debug(
            "Started with max queue size of %s events and metrics interval of %s seconds." % (
                self.config.size,
                self.config.frequency
            )
        )
        self.stopped = False

        if not self.name.startswith("_"):
            self.logging.debug("Started version %s" % (self.version))

    def sendToBackground(self, function, *args, **kwargs):
        '''

        Executes a function and sends it to the background. Such a function
        should never exit until ``self.loop`` returns ``False``.
        This `method` wraps ``function`` again in a loop as long ``self.loop``
        returns ``False`` so that ``function`` is restarted and an error is
        logged.

        Args:
            function (``function``): The function which has to be executed.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        '''

        def wrapIntoLoop():
            while self.loop():
                try:
                    function(*args, **kwargs)
                    # We want to break out of the loop if we get here because
                    # it's the intention of function() to exit without errors.
                    # Normally, background tasks run indefinately but in this
                    # case the user opted not to for some reason so we should
                    # obey that.
                    break
                except Exception as err:
                    if self.config.disable_exception_handling:
                        raise
                    self.logging.error("Backgrounded function '%s' of module instance '%s' caused an error. This needs attention. Restarting it in 2 seconds. Reason: %s" % (
                        function.__name__,
                        self.name,
                        err)
                    )
                    sleep(2)

        self.greenlets.generic.append(spawn(wrapIntoLoop))

    def stop(self):
        '''
        Makes ``self.loop`` return ``False`` and handles shutdown of of the
        registered background jobs.
        '''

        self.logging.info("Received stop. Initiating shutdown.")

        self.__loop = False

        for background_job in self.greenlets.metric:
            kill(background_job)

        for background_job in self.greenlets.generic:
            kill(background_job)

        for background_job in self.greenlets.consumer:
            kill(background_job)

        if hasattr(self, "postHook"):
            self.logging.debug("postHook() found, executing")
            self.postHook()

        self.logging.debug("Exit.")

        self.stopped = True

    def submit(self, event, queue):
        '''
        Submits <event> to the queue with name <queue>.


        Args:
            event (wishbone.event.Event): An event instance.
            queue (str): The name of the queue

        Returns:
            None
        '''

        while self.loop():
            try:
                getattr(self.pool.queue, queue).put(event)
                break
            except AttributeError:
                self.logging.error("No such queue %s. Event with uuid %s dropped." % (queue, event.get('uuid')))
                break
            except QueueFull:
                self.logging.warning("Queue '%s' is full and stalls the event pipeline. You should probably look into this." % (queue))
                sleep(0.1)

    def _applyFunctions(self, queue, event):
        '''
        Executes and applies all registered module functions against the event.

        Args:
            queue (str): The name of the queue to which the function was registered.
            event (wishbone.event.Event): The Wishbone

        Returns:
            wisbone.event.Event: The modified version of ``event``
        '''

        if queue in self.config.module_functions:
            for f in self.config.module_functions[queue]:
                try:
                    event = f.do(event)
                except Exception as err:
                    if self.config.disable_exception_handling:
                        raise
                    self.logging.error("Function '%s' is skipped as it is causing an error. Reason: '%s'" % (f, err))
        return event

    def _consumer(self, function, queue):
        '''
        Greenthread which applies <function> to each element from <queue>

        Args:
            function (``function``): The function which has been registered to consume ``queue``.

            queue (str): The name of the queue from which events have to be
                         consumed and processed by ``function``.

        Returns:
            None
        '''

        self._run.wait()
        self.logging.debug("Function '%s' has been registered to consume queue '%s'" % (function.__name__, queue))

        while self.loop():

            event = self.pool.getQueue(queue).get()
            if not event.has("tmp.%s" % (self.name)):
                event.set({}, "tmp.%s" % (self.name))

            # Render kwargs relative to the event's content and make these accessible under event.kwargs
            event.renderKwargs(self.kwargs_template)

            # Validate TTL
            try:
                event.decrementTTL()
            except TTLExpired as err:
                self.logging.warning("Event with UUID %s dropped. Reason: %s" % (event.get("uuid"), err))
                continue

            # Set the current event uuid to the logger object
            self.logging.setCurrentEventID(event.get("uuid"))

            # Apply all the defined queue functions to the event
            event = self._applyFunctions(queue, event)

            # Apply consumer function
            try:
                function(event)
            except Exception as err:
                if self.config.disable_exception_handling:
                    raise
                exc_type, exc_value, exc_traceback = exc_info()
                info = (traceback.extract_tb(exc_traceback)[-1][1], str(exc_type), str(exc_value))

                event.set(info, "errors.%s" % (self.name))

                self.logging.error("%s" % (err))
                self.submit(event, "_failed")
            else:
                self.submit(event, "_success")
            finally:
                # Unset the current event uuid to the logger object
                self.logging.setCurrentEventID(None)

    def __getDescription(self, config):
        '''
        Gets the module description.

        Args:
            config (``wishbone.actorconfig.ActorConfig``): An ActorConfig instance

        Returns:
            str: The description of this actor instance.
        '''

        if config.description is None:
            return self.__doc__.strip().split('\n')[0].strip('*')
        else:
            return config.description

    def __getRawKwargs(self):
        '''
        Get the class paramaters of the class basing this class.

        Returns (dict): A dict of the the raw kwargs
        '''
        kwargs = {}
        for key, value in list(inspect.getouterframes(inspect.currentframe())[2][0].f_locals.items()):
            if key == "self" or isinstance(value, ActorConfig):
                next
            else:
                kwargs[key] = value
        return kwargs

    def __getTemplateKwargs(self, template_env, kwargs):
        '''
        Recurses through ``kwargs`` and returns a version of it in which all
        strings are replaced by jinja2 template instances.

        Args:
            template_env (Jinja2.Environment instance): The Jinja2 environment instance to
                                                        derive templates from.
            kwargs (dict): The dict of keyword/arguments.

        '''

        def recurse(data):

            if isinstance(data, str):
                try:
                    if len(list(template_env.parse(data).find_all(jinja2.nodes.Name))) > 0:
                        t = template_env.from_string(data)
                        return t
                    else:
                        return data
                except Exception as err:
                    return data

            elif isinstance(data, dict):
                for key, value in data.items():
                    data[key] = recurse(value)
                return data
            elif isinstance(data, list):
                for index, value in enumerate(data):
                    data[index] = recurse(value)
                return data
            else:
                return data

        return recurse(deepcopy(kwargs))

    def __getVersion(self):

        try:
            return get_distribution(self.__module__.split('.')[0]).version
        except Exception as err:
            return "unknown"

    def __metricProducer(self):
        '''
        A greenthread collecting the queue metrics at the defined interval.
        '''

        self._run.wait()
        hostname = socket.gethostname()
        while self.loop():
            for queue in self.pool.listQueues(names=True):
                for metric, value in list(self.pool.getQueue(queue).stats().items()):
                    event = Wishbone_Event({
                        "time": time(),
                        "type": "wishbone",
                        "source": hostname,
                        "name": "module.%s.queue.%s.%s" % (self.name, queue, metric),
                        "value": value,
                        "unit": "",
                        "tags": ()
                    })
                    self.submit(event, "_metrics")
            sleep(self.config.frequency)

    def __postHook(self):
        '''
        Is always executed when the module starts.
        '''

        self.logging.debug("Following template functions are available: %s" % ", ".join(
            self.config.template_functions.keys()
        )
        )

    def __renderTemplateKwargs(self, kwargs):

        def recurse(data):

            if isinstance(data, jinja2.environment.Template):
                try:
                    return data.render()
                except Exception as err:
                    return "#error: %s#" % (err)
                    # return self.template_strings[data]
            elif isinstance(data, dict):
                result = {}
                for key, value in data.items():
                    result[key] = recurse(value)
                return EasyDict(result)
            elif isinstance(data, list):
                result = []
                for value in data:
                    result.append(recurse(value))
                return result
            else:
                return data

        rendered_kwargs = EasyDict(
            recurse(
                kwargs
            )
        )

        return rendered_kwargs

    def __validateAppliedFunctions(self):
        '''
        A validation routine which checks whether functions have been applied
        to queues without a registered consumer.  The effect of that would be
        that the functions are never applied which is not what the user
        wanted.
        '''

        queues_w_registered_consumers = [t.args[1] for t in self.greenlets.consumer]

        for queue in self.config.module_functions.keys():
            if queue not in queues_w_registered_consumers:
                raise ModuleInitFailure("Failed to initialize module '%s'. You have functions defined on queue '%s' which doesn't have a registered consumer." % (self.name, queue))

    def __sanityChecks(self):
        '''
        Does following validations:

            - Validate if all template functions base ``TemplateFunction``
            - Validate if all module functions base ``ModuleFunction``
            - Validate if the module has attribute "MODULE_TYPE" indicating it's not an pre 3.0 module.

        Args:
            None

        Returns:
            None

        Raises:
            ModuleInitFailure: Raised when one of the components isn't correct.

        '''

        # Validate template functions
        for n, f in self.config.template_functions.items():
            if not isinstance(f, TemplateFunction):
                raise ModuleInitFailure("Template function '%s' does not base TemplateFunction." % (n))

        # Validate module functions
        for name, functions in self.config.module_functions.items():

            for function in functions:
                if not isinstance(function, ModuleFunction):
                    raise ModuleInitFailure("Module function '%s' does not base ModuleFunction." % (name))

        if not hasattr(self, "MODULE_TYPE"):
            raise InvalidModule("Module instance '%s' seems to be of an incompatible old type." % (self.name))



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  logging.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone.event import Event
from wishbone.error import QueueFull
from time import time
from os import getpid


class MockLogger():
    '''
    A wrapper around Logging which mimics logger.Logger
    '''

    def __init__(self, name, q, level=5):

        self.level = level
        self.l = Logging(name, q)

    def flush(self):
        pass

    def write(self, line):
        self.l._Logging__log(self.level, line.rstrip())

    def writelines(self, lines):
        for line in lines:
            self.l.Logging.__log(self.level, line.rstrip())


class Logging():

    '''
    Generates Wishbone formatted log messages following the Syslog priority
    definition.
    '''

    LEVELS = {
        0: "emergency",
        1: "alert",
        2: "critical",
        3: "error",
        4: "warning",
        5: "notice",
        6: "informational",
        7: "debug"
    }

    def __init__(self, name, q, identification=None):
        self.name = name
        self.logs = q
        self.identification = identification
        self.__queue_full_message = False
        self.__event_id = None

    def __log(self, level, message):

        event = Event({
            "time": time(),
            "identification": self.identification,
            "event_id": self.__event_id,
            "level": level,
            "txt_level": self.LEVELS[level],
            "pid": getpid(),
            "module": self.name,
            "message": message
        })

        try:
            self.logs.put(event)
        except QueueFull:
            if not self.__queue_full_message:
                print("Log queue full for module '%s'. Dropping messages" % (self.name))
            else:
                self.__queue_full_message = True

    def alert(self, message, *args, **kwargs):
        """Generates a log message with priority alert(1).
        """
        self.__log(1, message)

    def critical(self, message, *args, **kwargs):
        """Generates a log message with priority critical(2).
        """
        self.__log(2, message)
    crit = critical

    def debug(self, message, *args, **kwargs):
        """Generates a log message with priority debug(7).
        """
        self.__log(7, message)

    def emergency(self, message, *args, **kwargs):
        """Generates a log message with priority emergency(0).
        """
        self.__log(0, message)
    emerg = emergency
    exception = emergency

    def error(self, message, *args, **kwargs):
        """Generates a log message with priority error(3).
        """
        self.__log(3, message, *args, **kwargs)
    err = error

    def informational(self, message, *args, **kwargs):
        """Generates a log message with priority informational(6).
        """
        self.__log(6, message)
    info = informational

    def notice(self, message, *args, **kwargs):
        """Generates a log message with priority notice(5).
        """
        self.__log(5, message)

    def setCurrentEventID(self, event_id):

        self.__event_id = event_id

    def warning(self, message, *args, **kwargs):
        """Generates a log message with priority warning(4).
        """
        self.__log(4, message, *args, **kwargs)
    warn = warning



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  generator.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import InputModule
from gevent import sleep


class Generator(InputModule):
    '''
    Generates an event at the chosen interval.

    The payload can be just about anything including template functions.


    Parameters::

        - destination(str)("data")
           |  The location write the payload to

        - interval(float)(1)
           |  The interval in seconds between each generated event.
           |  A value of 0 means as fast as possible.

        - native_events(bool)(False)
           |  Whether to expect incoming events to be native Wishbone events

        - payload(str/dict/int/float)("test")
           |  The content of the test message.

    Queues::

        - outbox
           |  Contains the generated events.
    '''

    def __init__(self, actor_config, native_events=False,
                 interval=1, payload="test", destination="data"):
        InputModule.__init__(self, actor_config)
        self.pool.createQueue("outbox")

    def preHook(self):

        self.sendToBackground(self.produce)

    def produce(self):

        while self.loop():
            for chunk in [self.kwargs_raw["payload"], None]:
                for payload in self.decode(chunk):
                    event = self.generateEvent(
                        payload,
                        self.kwargs.destination
                    )
                    self.submit(event, "outbox")
            sleep(self.kwargs.interval)
        self.logging.info("Stopped producing events.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  random_word.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.function.template import TemplateFunction
import os
from random import choice
import io


class RandomWord(TemplateFunction):

    '''
    Returns a random word.

    A Wishbone template function which returns a randomly selected word from a
    word list.

    Args:

        filename (str): When None, a built-in word list is used otherwise the
                        referenced file is loaded.
        encoding (str): The encoding used to read the file.
    '''

    def __init__(self, filename=None, encoding="latin-1"):

        if filename is None:
            self.filename = "%s/../../data/wordlist.txt" % (os.path.dirname(__file__))
        else:
            self.filename = filename

        self.wordlist = self.readWordlist(filename, encoding)

    def readWordlist(self, location, encoding):
        with io.open(self.filename, encoding=encoding) as f:
            return f.readlines()

    def get(self):
        '''
        The function mapped to the template function.

        Args:
            None

        Returns:
            str: A random word
        '''

        return choice(self.wordlist).rstrip()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  fileout.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone.actor import Actor
from wishbone.module import OutputModule
from gevent.os import make_nonblocking
import arrow


class FileOut(OutputModule):

    '''**Writes events to a file**

    Writes incoming events to a file.  Each line represents an event. Keep in
    mind no rotation of the file is done so data is always appended to the end
    of the file.


    Parameters::

        - selection(str)("data")
           |  The part of the event to submit externally.
           |  Use an empty string to refer to the complete event.

        - location(str)("./wishbone.out")
           |  The location of the output file.

        - timestamp(bool)(False)
           |  If true prepends each line with a ISO8601 timestamp.

    Queues::

        - inbox
           |  Incoming messages

    '''

    def __init__(self, actor_config, selection='data', location="./wishbone.out", timestamp=False):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        if self.kwargs.timestamp:
            self.getTimestamp = self.returnTimestamp
        else:
            self.getTimestamp = self.returnNoTimestamp

        self.file = open(self.kwargs.location, "a")
        make_nonblocking(self.file)

    def consume(self, event):

        if event.isBulk():
            data = event.dumpFieldAsString(self.kwargs.selection)
        else:
            data = str(event.get(self.kwargs.selection))

        self.file.write("%s%s\n" % (self.getTimestamp(), data))
        self.file.flush()

    def returnTimestamp(self):

        return "%s: " % (arrow.now().isoformat())

    def returnNoTimestamp(self):

        return ""

    def postHook(self):

        self.file.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone.actor import Actor
from wishbone.moduletype import ModuleType
from wishbone.componentmanager import ComponentManager
from wishbone.error import ModuleInitFailure
from wishbone.event import extractBulkItemValues
from wishbone.protocol import Encode, Decode
from wishbone.event import Event as Wishbone_Event
from wishbone.protocol.decode.dummy import Dummy as DummyDecoder
from wishbone.protocol.encode.dummy import Dummy as DummyEncoder
from wishbone.error import TTLExpired
from copy import deepcopy
from sys import exc_info
import traceback
from gevent.pool import Pool


class InputModule(Actor):
    MODULE_TYPE = ModuleType.INPUT

    def setDecoder(self, name, *args, **kwargs):
        '''
        Sets the decoder with name <name> unless there's already a decoder
        defined via ``actorconfig.ActorConfig``.

        Args:
            name (str): The name of the decoder to initialize
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Bool: True if the decoder is set, False when a decoder was already
                  set via ``actorconfig.ActorConfig``
        '''
        if self.decode is None:
            self.decode = ComponentManager().getComponentByName(name)(*args, **kwargs).handler
            self.logging.debug("Decoder '%s' has been set.")
            return True
        else:
            self.logging.debug("Decoder '%s' has not been set. The user already defined one." % (name))
            return False

    decode = DummyDecoder().handler

    def _generateNativeEvent(self, data={}, destination=None):
        '''
        Gets mapps to self.generateEvent for `input` type modules and if
        ``Actor.config.protocol_event`` is ``True``.

        Args:
            data (dict): The dict representation of ``wishbone.event.Event``.
            destination (str): The destination str to store ``data``. In this
                               particular implementation it's ignored.

        Returns:
            wishbone.event.Event: ``Event`` instance of ``data``
        '''

        e = Wishbone_Event()
        e.slurp(data)
        return e

    def _moduleInitSetup(self):
        '''
        Does module type specific setup.
        '''

        if not hasattr(self, "decode") and self.config.protocol is None:
            self.logging.debug("This 'Input' module has no decoder method set. Setting dummy decoder.")
            self.setDecoder("wishbone.protocol.decode.dummy")
        if self.config.protocol is not None:
            self.logging.debug("This 'Input' module has '%s' decoder configured." % (self.config.protocol))
            self.decode = self.config.protocol.handler

        if self.kwargs.native_events:
            self.generateEvent = self._generateNativeEvent

    def _moduleInitValidation(self):
        '''
        Validates whether we have all the parameters this module type expects

        Args:
            None

        Returns:
            None

        Raises:
            ModuleInitFailure: Raised when one of the components isn't correct.
        '''

        if self.config.protocol is not None and not isinstance(self.config.protocol, Decode):
            raise ModuleInitFailure("An 'input' module should have a decode protocol set. Found %s" % (type(self.config.protocol)))

        for param in ["native_events", "destination"]:
            if param not in self.kwargs.keys():
                raise ModuleInitFailure("An 'Input' module should always have a '%s' parameter. This is a programming error." % (param))


class OutputModule(Actor):
    MODULE_TYPE = ModuleType.OUTPUT

    def setEncoder(self, name, *args, **kwargs):
        '''
        Sets the encoder with name <name> unless there's already an encoder
        defined via ``actorconfig.ActorConfig``.

        Args:
            name (str): The name of the encoder to initialize
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Bool: True if the encoder is set, False when an encoder was already
                  set via ``actorconfig.ActorConfig``
        '''

        if self.encode is None:
            self.encode = ComponentManager().getComponentByName(name)(*args, **kwargs).handler
            self.logging.debug("Encoder '%s' has been set.")
            return True
        else:
            self.logging.debug("Encoder '%s' has not been set. The user already defined one." % (name))
            return False

    encode = DummyEncoder().handler

    def getDataToSubmit(self, event):
        '''
        Derives the data to submit from ``event`` taking into account
        ``native_events``, ``payload`` and ``selection`` module parameters.

        Args:
            event (```wishbone.event.Event```): The event to extract data from.

        Returns:

            dict/str/...: The data to submit.
        '''

        if self.kwargs.native_events:
            return event.getNative()
        elif event.kwargs.payload is None:
            if event.isBulk():
                return "\n".join([str(item) for item in extractBulkItemValues(event, self.kwargs.selection)])
            else:
                return event.get(
                    event.kwargs.selection
                )
        else:
            return event.kwargs.payload

    def _consumer(self, function, queue):
        '''

        Greenthread which applies <function> to each element from <queue>.
        However, this version overrides ``Actor._consumer`` as it executes
        parallel coroutine versions of ``functions`` spawned on a
        ``gevent.pool.Pool`` instance.  The number of parallel instances is
        defined by the ``parallel_streams`` value.

        Args:
            function (``function``): The function which has been registered to consume ``queue``.

            queue (str): The name of the queue from which events have to be
                         consumed and processed by ``function``.

        Returns:
            None
        '''

        self.parallel_pool = Pool(self.kwargs.parallel_streams)
        self._run.wait()
        self.logging.debug("Function '%s' has been registered to consume queue '%s'" % (function.__name__, queue))

        def execFunction(function, event):
            try:
                function(deepcopy(event))
            except Exception as err:
                if self.config.disable_exception_handling:
                    raise
                exc_type, exc_value, exc_traceback = exc_info()
                info = (traceback.extract_tb(exc_traceback)[-1][1], str(exc_type), str(exc_value))

                event.set(info, "errors.%s" % (self.name))

                self.logging.error("%s" % (err))
                self.submit(event, "_failed")
            else:
                self.submit(event, "_success")
            finally:
                # Unset the current event uuid to the logger object
                self.logging.setCurrentEventID(None)

        while self.loop():

            event = self.pool.getQueue(queue).get()
            if not event.has("tmp.%s" % (self.name)):
                event.set({}, "tmp.%s" % (self.name))

            # Render kwargs relative to the event's content and make these accessible under event.kwargs
            event.renderKwargs(self.kwargs_template)

            # Validate TTL
            try:
                event.decrementTTL()
            except TTLExpired as err:
                self.logging.warning("Event with UUID %s dropped. Reason: %s" % (event.get("uuid"), err))
                continue

            # Set the current event uuid to the logger object
            self.logging.setCurrentEventID(event.get("uuid"))

            # Apply all the defined queue functions to the event
            event = self._applyFunctions(queue, event)

            # Apply consumer function
            self.parallel_pool.spawn(execFunction, function, event)

    def _moduleInitSetup(self):
        '''
        Does module type specific setup.
        '''

        if not hasattr(self, "encode") and self.config.protocol is None:
            self.logging.debug("This 'Output' module has no encoder method set. Setting dummy decoder.")
            self.setEncoder("wishbone.protocol.encode.dummy")
        if self.config.protocol is not None:
            self.logging.debug("This 'Output' module has '%s' encoder configured." % (self.config.protocol))
            self.encode = self.config.protocol.handler

    def _moduleInitValidation(self):
        '''
        Validates whether we have all the parameters this module type expects

        Args:
            None

        Returns:
            None

        Raises:
            ModuleInitFailure: Raised when one of the components isn't correct.
        '''

        if self.config.protocol is not None and not isinstance(self.config.protocol, Encode):
            raise ModuleInitFailure("An 'output' module should have a encode protocol set. Found %s" % (type(self.config.protocol)))

        for param in ["payload", "selection", "native_events", "parallel_streams"]:
            if param not in self.kwargs.keys():
                raise ModuleInitFailure("An 'output' module should always have a '%s' parameter. This is a programming error." % (param))


class FlowModule(Actor):
    MODULE_TYPE = ModuleType.FLOW

    def _moduleInitSetup(self):
        '''
        Does module type specific setup.
        '''

        pass

    def _moduleInitValidation(self):
        '''
        Validates whether we have all the parameters this module type expects

        Args:
            None

        Returns:
            None

        Raises:
            ModuleInitFailure: Raised when one of the components isn't correct.
        '''

        pass


class ProcessModule(Actor):
    MODULE_TYPE = ModuleType.PROCESS

    def _moduleInitSetup(self):
        '''
        Does module type specific setup.
        '''

        pass

    def _moduleInitValidation(self):
        '''
        Validates whether we have all the parameters this module type expects

        Args:
            None

        Returns:
            None

        Raises:
            ModuleInitFailure: Raised when one of the components isn't correct.
        '''

        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  queue.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from uuid import uuid4
from gevent.queue import Queue as Gevent_Queue
from wishbone.error import ReservedName, QueueMissing, QueueFull, QueueEmpty
from time import time
from gevent.queue import Empty, Full
from gevent import sleep
from types import SimpleNamespace


class QueuePool():

    RESERVED_QUEUES = [
        '_failed',
        '_success',
        '_logs',
        '_metrics'
    ]

    def __init__(self, size):
        self.__size = size
        self.queue = SimpleNamespace()
        self.queue._metrics = Queue(size)
        self.queue._logs = Queue(size)
        self.queue._success = Queue(size)
        self.queue._failed = Queue(size)

    def listQueues(self, names=False, default=True):
        '''returns the list of queue names from the queuepool.
        '''

        if default:
            blacklist = []
        else:
            blacklist = self.RESERVED_QUEUES

        for m in list(self.queue.__dict__.keys()):
            if m not in blacklist:
                if not names:
                    yield getattr(self.queue, m)
                else:
                    yield m

    def createQueue(self, name):
        '''Creates a Queue.'''

        if name in self.RESERVED_QUEUES or name.startswith("_"):
            raise ReservedName("%s is an invalid queue name." % (name))

        setattr(self.queue, name, Queue(self.__size))

    def createSystemQueue(self, name):
        '''
        Creates a queue starting with _ which is forbidden by users as special
        system queues are prefixed with '_'.

        This method should normally never be used directly because of user
        input.  It's the Wishbone framework itself which should make system
        related queues.

        Args:
            name (str): The name of the queue (should start with _)
        '''

        setattr(self.queue, name, Queue(self.__size))

    def hasQueue(self, name):
        '''Returns <True> when queue with <name> exists.'''

        try:
            getattr(self.queue, name)
            return True
        except:
            return False

    def getQueue(self, name):
        '''Convenience funtion which returns the queue instance.'''

        try:
            return getattr(self.queue, name)
        except:
            raise QueueMissing("Module has no queue %s." % (name))

    def join(self):
        '''Blocks until all queues in the pool are empty.'''
        counter = 0
        for queue in self.listQueues():
            while queue.qsize() != 0:
                sleep(0.2)
                counter += 1
                if counter == 5:
                    break


class Queue():

    '''A queue used to organize communication messaging between Wishbone Actors.

    Parameters:

        - max_size (int):   The max number of elements in the queue.
                            Default: 1

    When a queue is created, it will drop all messages. This is by design.
    When <disableFallThrough()> is called, the queue will keep submitted
    messages.  The motivation for this is that when is queue is not connected
    to any consumer it would just sit there filled and possibly blocking the
    chain.

    The <stats()> function will reveal whether any events have disappeared via
    this queue.

    '''

    def __init__(self, max_size=1):
        self.max_size = max_size
        self.id = str(uuid4())
        self.__q = Gevent_Queue(max_size)
        self.__in = 0
        self.__out = 0
        self.__dropped = 0
        self.__cache = {}

        self.put = self.__fallThrough

    def clean(self):
        '''Deletes the content of the queue.
        '''
        self.__q = Gevent_Queue(self.max_size)

    def disableFallThrough(self):
        self.put = self.__put

    def dump(self):
        '''Dumps the queue as a generator and cleans it when done.
        '''

        while True:
            try:
                yield self.get(block=False)
            except QueueEmpty:
                break

    def empty(self):
        '''Returns True when queue and unacknowledged is empty otherwise False.'''

        return self.__q.empty()

    def enableFallthrough(self):
        self.put = self.__fallThrough

    def get(self, block=True):
        '''Gets an element from the queue.'''

        try:
            e = self.__q.get(block=block)
        except Empty:
            raise QueueEmpty("Queue is empty.")
        self.__out += 1
        return e

    def rescue(self, element):

        self.__q.put(element)

    def size(self):
        '''Returns the length of the queue.'''

        return self.__q.qsize()

    def stats(self):
        '''Returns statistics of the queue.'''

        return {"size": self.__q.qsize(),
                "in_total": self.__in,
                "out_total": self.__out,
                "in_rate": self.__rate("in_rate", self.__in),
                "out_rate": self.__rate("out_rate", self.__out),
                "dropped_total": self.__dropped,
                "dropped_rate": self.__rate("dropped_rate", self.__dropped)
                }

    def __fallThrough(self, element):
        '''Accepts an element but discards it'''

        self.__dropped += 1
        del(element)

    def __put(self, element):
        '''Puts element in queue.'''

        try:
            self.__q.put(element)
            self.__in += 1
        except Full:
            raise QueueFull("Queue full.")

    def __rate(self, name, value):

        if name not in self.__cache:
            self.__cache[name] = {"value": (time(), value), "rate": 0}
            return 0

        (time_then, amount_then) = self.__cache[name]["value"]
        (time_now, amount_now) = time(), value

        if time_now - time_then >= 1:
            self.__cache[name]["value"] = (time_now, amount_now)
            self.__cache[name]["rate"] = (amount_now - amount_then) / (time_now - time_then)

        return self.__cache[name]["rate"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  acknowledge.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.module import FlowModule
from gevent.lock import Semaphore
from random import SystemRandom
import string


class AckList(object):

    def __init__(self):

        self.ack_table = []
        self.lock = Semaphore()

    def ack(self, value):

        with self.lock:
            if value in self.ack_table:
                self.ack_table.remove(value)
                return True
            else:
                return False

    def unack(self, value):

        with self.lock:
            if value in self.ack_table:
                return False
            else:
                self.ack_table.append(value)
                return True


class Acknowledge(FlowModule):

    '''**Forwards or drops events by acknowleding values.**

    This module stores the value of field ``ack_id`` from each incoming event.
    Subsequent events with the same ``ack_id`` field value will be dropped
    until it is removed by having it acknowledged.

    The ``ack_id`` can be acknowledged by sending the event to the
    ``acknowledge`` queue.

    The ``ack_id`` field value should be an unique value.

    Typically, downstream modules's ``successful`` and/or ``failed`` queues are
    sending events to the ``acknowledge`` queue.


    Parameters::

        - ack_id (data):
            A unique value identifying the event.

    Queues::

        - inbox
            |  Incoming events

        - outbox
            |  Outgoing events

        - acknowledge
            |  Acknowledge events

        - dropped
            |  Where events go to when unacknowledged


    Event variables::

        - tmp.<name>.ack_id
            | The location of the acknowledgement ID when coming in through the
            | inbox queue.
    '''

    def __init__(self, actor_config, ack_id=None):
        FlowModule.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue("acknowledge")
        self.pool.createQueue("dropped")
        self.registerConsumer(self.consume, "inbox")
        self.registerConsumer(self.acknowledge, "acknowledge")

        self.ack_table = AckList()

    def consume(self, event):

        if self.kwargs.ack_id is None:
            ack_id = self.generateID()
        else:
            ack_id = event.kwargs.ack_id

        if event.has("tmp.%s.ack_id" % (self.name)):
            self.logging.warning("Event arriving to <inbox> with tmp.%s.ack_id already set.  Perhaps that should have been the <acknowledge> queue instead." % (self.name))
        else:
            event.set({"ack_id": ack_id}, "tmp.%s" % (self.name))
            if self.ack_table.unack(ack_id):
                self.submit(event, "outbox")
            else:
                self.logging.debug("Event with still unacknowledged <ack_id> '%s' send to <dropped> queue." % (ack_id))
                self.submit(event, "dropped")

    def acknowledge(self, event):

        if event.has("tmp.%s.ack_id" % (self.name)):
            ack_id = event.get('tmp.%s.ack_id' % (self.name))
            if self.ack_table.ack(ack_id):
                self.logging.debug("Event acknowledged with <ack_id> '%s'." % (ack_id))
                event.delete('tmp.%s.ack_id' % (self.name))
            else:
                self.logging.debug("Event with <ack_id> '%s' received but was not previously acknowledged." % (ack_id))
        else:
            self.logging.warning("Received event without 'tmp.%s.ack_id' therefor it is dropped" % (self.name))

    def generateID(self):

        return ''.join(SystemRandom().choice(string.ascii_lowercase) for _ in range(4))

    def postHook(self):

        self.logging.debug("The ack table has %s events unacknowledged." % (len(self.ack_table.ack_table)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  ttl.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor


class TTL(Actor):

    '''**Allows messages to pass a maximum number of times.**

    When a message has traveled through this module more than <ttl> times it
    will be submitted to the <ttl_exceeded> queue.

    Parameters::

        - ttl(int)(1)
           |  The maximum number of times an event is allowed
           |  to travel through.


    Queues::

        - inbox
           |  Incoming events.

        - outbox
           |  Outgoing events.

        - ttl_exceeded
           |  Events which passed the module more than <ttl> times.
    '''

    def __init__(self, actor_config, ttl=1):
        Actor.__init__(self, actor_config)

        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")
        self.pool.createQueue("outbox")
        self.pool.createQueue("ttl_exceeded")

        self.__flush_lock = False

    def consume(self, event):

        if self.validateTTL(event):
            self.submit(event, self.pool.queue.outbox)
        else:
            self.logging.warning("Event TTL of %s exceeded in transit (%s) moving event to ttl_exceeded queue." % (event.getHeaderValue(self.name, "ttl_counter"), self.kwargs.ttl))
            self.submit(event, self.pool.queue.ttl_exceeded)

    def validateTTL(self, event):

        try:
            value = event.getHeaderValue(self.name, "ttl_counter")
            value += 1
            event.setHeaderValue("ttl_counter", value)
        except KeyError:
            event.setHeaderValue("ttl_counter", 1)
            value = 1

        if value > self.kwargs.ttl:
            return False
        else:
            return True



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  py2devtools.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#
# The MIT License (MIT)

# Copyright (c) 2014 Nylas

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Original source https://github.com/nylas/nylas-perftools/blob/master/py2devtools.py


import json
import sys
import timeit
import gevent
from os import getpid


class Node(object):
    def __init__(self, name, id_):
        self.name = name
        self.id_ = id_
        self.children = {}
        self.hitCount = 1

    def serialize(self):
        res = {
            'functionName': self.name,
            'hitCount': self.hitCount,
            'children': [c.serialize() for c in list(self.children.values())],
            'scriptId': '1',
            'url': '',
            'lineNumber': 1,
            'columnNumber': 1,
            'deoptReason': '',
            'id': self.id_,
            'callUID': self.id_
        }
        return res

    def add(self, frames, idgen):
        if not frames:
            self.hitCount += 1
            return
        head = frames[0]
        child = self.children.get(head)
        if child is None:
            child = Node(name=head, id_=idgen())
            self.children[head] = child
        child.add(frames[1:], idgen)


class Profiler(object):
    def __init__(self, target_greenlet=None, interval=0.0001):
        if target_greenlet is not None:
            self.target_greenlet_id = id(target_greenlet)
        else:
            self.target_greenlet_id = None
        self.interval = interval
        self.started = None
        self.last_profile = None
        self.root = Node('head', 1)
        self.nextId = 1
        self.samples = []
        self.timestamps = []

    def _idgenerator(self):
        self.nextId += 1
        return self.nextId

    def _profile(self, frame, event, arg):
        if event == 'call':
            self._record_frame(frame.f_back)

    def _record_frame(self, frame):
        if (self.target_greenlet_id is None or
                id(gevent.getcurrent()) == self.target_greenlet_id):
            now = timeit.default_timer()
            if self.last_profile is not None:
                if now - self.last_profile < self.interval:
                    return
            self.last_profile = now
            self.timestamps.append(int(1e6 * now))
            stack = []
            while frame is not None:
                stack.append(self._format_frame(frame))
                frame = frame.f_back
            stack.reverse()
            self.root.add(stack, self._idgenerator)
            self.samples.append(self.nextId)

    def _format_frame(self, frame):
        return '{}({})'.format(frame.f_code.co_name,
                               frame.f_globals.get('__name__'))

    def output(self):
        if not self.samples:
            return {}
        return json.dumps({
            'startTime': self.started,
            'endTime': 0.000001 * self.timestamps[-1],
            'timestamps': self.timestamps,
            'samples': self.samples,
            'head': self.root.serialize()
        })

    def __enter__(self):
        sys.setprofile(self._profile)
        self.started = timeit.default_timer()
        print("")
        print("########################################")
        print("#                                      #")
        print("# Wishbone is running in profile mode. #")
        print("# This has a performance impact.       #")
        print("#                                      #")
        print("########################################")
        print("")

    def __exit__(self, type, value, traceback):
        sys.setprofile(None)
        filename = './wishbone_%s_.cpuprofile' % getpid()
        with open(filename, 'w') as f:
            f.write(self.output())
            print(("Written profile file '%s'." % (filename)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  cron.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from gevent import monkey; monkey.patch_all()
from wishbone.actor import Actor
from wishbone.module import InputModule
from cronex import CronExpression
from gevent import sleep
import time


class Cron(InputModule):

    '''**Generates an event at the defined time**

    Generates an event with the defined payload at the chosen time.
    Time is in crontab format.


    Parameters::

        - native_events(bool)(False)
           |  Whether to expect incoming events to be native Wishbone events

        - cron(string)("*/10 * * * *")
            | The cron expression.

        - payload(str)("wishbone")
            | The content of <destination>.

        - destination(str)("data")
            | The location to write <payload> to.


    Queues::

        - outbox
           |  Outgoing messges
    '''

    def __init__(self, actor_config, native_events=False,
                 cron="*/10 * * * *", payload="wishbone", destination="data"):

        Actor.__init__(self, actor_config)
        self.pool.createQueue("outbox")
        self.cron = CronExpression("%s wishbone" % self.kwargs.cron)

    def preHook(self):
        self.sendToBackground(self.timer)

    def timer(self):
        while self.loop():
            if self.cron.check_trigger(time.localtime(time.time())[:5]):
                self.logging.info("Cron executed.")
            for chunk in [self.kwargs_raw["payload"], None]:
                for payload in self.decode(chunk):
                    event = self.generateEvent(
                        payload,
                        self.kwargs.destination
                    )
                    self.submit(
                        event,
                        "outbox"
                    )

            sleep(60)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  modulemanager.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import pkg_resources
import re
from prettytable import PrettyTable
from wishbone.error import NoSuchModule, InvalidModule
from wishbone import Actor
from wishbone.lookup import Lookup


class ModuleManager():

    '''
    Manages the Wishbone modules it can find under the defined entrypoints
    prefixes.

    When initiated it indexes all the Wishbone actor modules it can find in
    the <categories>.<groups> combinations.

    Note:

        A module manager expects to find Wishbone actor based modules as
        entrypoints with prefixes like "wishbone.flow", "wishbone.encode" etc,
        ...*

        Wishbone categories are "wishbone" and "wishbone_contrib" where the
        first holds the "official" builtin modules whereas the latter should
        be the namespace to register communitry contributed (external)
        modules.*

        Each category is divided into 7 groups "flow", "encode", "decode",
        "function", "input", "output", "lookup" which define the type of
        module.*

    Args:

        categories (list): The list of categories to search for <groups>
        groups (list): The list of groups to search for Wishbone actor
            modules.

    '''

    def __init__(self,
                 categories=["wishbone", "wishbone_contrib"],
                 groups=["flow", "encode", "decode", "function", "input", "output", "lookup"]):
        self.categories = categories
        self.groups = groups

    def exists(self, name):

        '''
        Validates whether the module with <name> exists.

        Args:
            name (str): The complete name of the module.

        Returns:
            bool: True if module exists. False otherwise.
        '''

        if self.getModuleByName(name) is None:
            return True
        else:
            return False

    def getModule(self, category, group, name):
        '''
        Returns the module with name <category>.<group>.<name>

        Args:
            category (str): The category name.
            group (str): The group name.
            name (str): The module name.

        Returns:
            class: A ``wishbone.Actor``` or ``wishbone.Lookup`` based class

        Raises:
            NoSuchModule: The module does not exist.
            InvalidModule: There was module found but it was not deemed valid.
        '''

        m = None
        for module in pkg_resources.iter_entry_points("%s.%s" % (category, group)):
            if module.name == name:
                m = module.load()

        if m is None:
            raise NoSuchModule("Module %s.%s.%s cannot be found." % (category, group, name))
        else:
            if issubclass(m, Actor) or issubclass(m, Lookup):
                return m
            else:
                raise InvalidModule("'%s.%s.%s' is not a valid wishbone actor or lookup module." % (category, group, name))

    def getModuleByName(self, name):
        '''
        Returns the module with name <name>

        Args:
            name (str): The complete module name.

        Returns:
            class: A `wishbone.Actor` or `wishbone.Lookup` based class

        Raises:
            NoSuchModule: The module does not exist.
            InvalidModule: There was module found but it was not deemed valid.
        '''

        (category, group, name) = name.split('.')
        m = self.getModule(category, group, name)

        if issubclass(m, Actor) or issubclass(m, Lookup):
            return m
        else:
            raise InvalidModule("'%s.%s.%s' is not a valid wishbone actor or lookup module." % (category, group, name))

    def getModuleList(self):
        '''
        Finds and lists all the modules it can find under the defined
        <category>.<groups> combinations.

        Args:
            category (str): The category name.
            group (str): The group name.
            name (str): The module name.

        Yields:
            tuple: A 3 element tuple: (`category`, `group`, `module`)
        '''
        for category in self.categories:
            for group in self.groups:
                group_name = "%s.%s" % (category, group)
                groups = [m.name for m in pkg_resources.iter_entry_points(group=group_name)]
                for m in sorted(groups):
                    yield (category, group, m)

    def getModuleDoc(self, category, group, name):
        '''
        Returns the docstring of module `category`.`group`.`name`

        Args:
            category (str): The category name.
            group (str): The group name.
            name (str): The module name.

        Returns:
            str: The docstring of the module.

        Raises:
            InvalidModule: The docstring does not have the correct format.
        '''

        doc = self.getModule(category, group, name).__doc__
        doc = re.search('(\*\*.*?\*\*)(.*)', doc, re.DOTALL).group(2)
        if doc is None:
            raise InvalidModule("The module does not seem to have the expected docstring format.")
        else:
            return doc

    def getModuleTitle(self, category, group, name):
        '''
        Returns the title of the module `category`.`group`.`name` docstring.

        Args:
            category (str): The category name.
            group (str): The group name.
            name (str): The module name.

        Returns:
            str: The docstring/module title

        Raises:
            InvalidModule: The docstring does not have the correct format.
        '''

        doc = self.getModule(category, group, name).__doc__
        title = re.search('\*\*(.*?)\*\*(.*)', doc).group(1)
        if title is None:
            raise InvalidModule("The module does not seem to have the expected docstring format.")
        else:
            return title

    def getModuleTable(self):
        '''
        Returns an ascii table of all Wishbone modules with valid entry points.

        Returns:
            str: The ascii table containing all modules.
        '''

        table = self.__getModuleTable()

        category_header = None
        group_header = None
        all_items = list(self.getModuleList())

        for (category, group, module) in all_items:

            title = self.getModuleTitle(category, group, module)
            version = self.getModuleVersion(category, group, module)
            if category_header == category:
                category = ""
            else:
                category_header = category
            if group_header == group:
                group = ""
            else:
                table.add_row(["", "", "", "", ""])
                group_header = group
            table.add_row([category, group, module, version, title])

        table.add_row(["", "", "", "", ""])
        return table.get_string()

    def getModuleVersion(self, category, group, name):
        '''
        Returns the version of the module.

        Args:
            category (str): The category name.
            group (str): The group name.
            name (str): The module name.

        Returns:
            str: The version of the module.
        '''

        try:
            for module in pkg_resources.iter_entry_points("%s.%s" % (category, group)):
                if module.name == name:
                    return module.dist.version
        except:
            return "?"

    def validateModuleName(self, name):
        '''
        Validates a module reference name for the proper format.

        Args:
            name (str): The name to validate.

        Returns:
            bool: True when valid.  False when invalid.
        '''

        if len(name.split('.')) != 3:
            return False
        else:
            return True

    def __getModuleTable(self):
        '''
        Returns a skeleton ascii module table object
        '''

        t = PrettyTable(["Category", "Group", "Module", "Version", "Description"])
        t.align["Category"] = "l"
        t.align["Group"] = "l"
        t.align["Module"] = "l"
        t.align["Version"] = "r"
        t.align["Description"] = "l"

        return t



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
#  roundrobin.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone.actor import Actor
from wishbone.module import FlowModule
from itertools import cycle
from random import randint


class RoundRobin(FlowModule):

    '''**Round-robins incoming events to all connected queues.**

    Create a "1 to n" relationship between queues.  Events arriving in inbox
    are then submitted in a roundrobin (or randomized) fashion to the
    connected queues.  The outbox queue is non existent.

    Parameters::

        - randomize(bool)(False)
            |  Randomizes the queue selection instead of going round-robin
            |  over all queues.


    Queues::

        - inbox
           |  Incoming events
    '''

    def __init__(self, actor_config, randomize=False):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("inbox")
        self.registerConsumer(self.consume, "inbox")

    def preHook(self):

        self.destination_queues = []
        for queue in self.pool.listQueues(names=True):
            if queue not in ["_failed", "_success", "_metrics", "_logs"]:
                self.destination_queues.append(queue)

        if not self.kwargs.randomize:
            self.cycle = cycle(self.destination_queues)
            self.chooseQueue = self.__chooseNextQueue
        else:
            self.chooseQueue = self.__chooseRandomQueue

    def consume(self, event):
        queue = self.chooseQueue()
        self.submit(event, queue)

    def __chooseNextQueue(self):
        return next(self.cycle)

    def __chooseRandomQueue(self):
        index = randint(0, len(self.destination_queues) - 1)
        return self.destination_queues[index]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  json.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


from wishbone.error import ProtocolError
from wishbone.protocol import Decode
from json import loads
from io import StringIO
import os


class JSON(Decode):

    '''
    Decode a JSON string into a Python object.

    Parameters:

        - charset(str)('utf-8)
            |  The charset to use when decoding.

        - delimiter(str)(None)
            |  The dilimiter to use in a stream of Python docs.
            |  None expects the complete payload to be 1 json document.

        - buffer_size(int)(4096)
            |  The maximum number of bytes allowed to buffer.

    '''

    def __init__(self, charset='utf-8', delimiter=None, buffer_size=4096):

        self.charset = charset
        if isinstance(delimiter, bytes):
            self.delimiter = delimiter.decode(charset, 'strict')
        else:
            self.delimiter = delimiter
        self.buffer_size = buffer_size
        self.buffer = StringIO()

    def getBufferSize(self):

        cur_pos = self.buffer.tell()
        self.buffer.seek(0, os.SEEK_END)
        size = self.buffer.tell()
        self.buffer.seek(cur_pos)
        return size

    def handleBytes(self, data):

        for item in self.handler(data.decode(self.charset, "strict")):
            yield self.handler(item)

    def handleDict(self, data):

        return data

    def handleGenerator(self, data):

        for chunk in data:
            for item in self.handler(chunk):
                yield item

    def handleReadLinesMethod(self, data):

        for chunk in data.readlines():
            for item in self.handler(chunk):
                yield item

    def handleString(self, data):

        self.buffer.write(data)
        self.buffer.seek(0)

        if self.delimiter is not None and self.delimiter in self.buffer.getvalue():
            for item in self.buffer.getvalue().split(self.delimiter):
                if item == "":
                    next
                else:
                    try:
                        yield loads(item)
                    except Exception as err:
                        raise ProtocolError(err)
            self.buffer.truncate(0)
        else:
            self.buffer.seek(0, 2)

    def flush(self):

        if len(self.buffer.getvalue()) == 0:
            return
            yield
        else:
            try:
                yield loads(
                    self.buffer.getvalue()
                )
            except Exception as err:
                raise ProtocolError(err)
            else:
                self.resetBuffer()

    def resetBuffer(self):

        self.buffer = StringIO()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  actorconfig.py
#
#  Copyright 2018 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#


class ActorConfig(object):

    '''
    A configuration object pass to a Wishbone actor.

    This is a simple object which holds a set of attributes (with some sane
    defaults) a Wishbone Actor expects.

    Attributes:
        name (str): The name identifying the actor instance.
        size (int): The size of the Actor instance's queues.
        frequency (int): The time in seconds to generate metrics.
        template_functions (dict): A dictionary of template functions.
        description (str): A short free form discription of the actor instance.
        module_functions (dict): A dict of queue names containing an array of module_functions
        protocol (``wishbone.protocol.Encode``, ``wishbone.protocol.Encode``): A protocol decode or encode instance.
        io_event (bool): When ``True`` Input and Output modules know to expect or emit serialzed wishbone events.
        identification (str): A name assigned to the Wishbone instance, useful for the module to know such as logging.
        disable_exception_handling (bool): If True, exception handling is disabled. Usefull for testing
    '''

    def __init__(self, name, size=100, frequency=10, template_functions={}, description=None, module_functions={},
                 protocol=None, io_event=False,
                 identification="wishbone",
                 disable_exception_handling=False):
        '''
        Args:
            name (str): The name identifying the actor instance.
            size (int): The size of the Actor instance's queues.
            frequency (int): The time in seconds to generate metrics.
            template_functions (dict): A dictionary of template functions.
            description (str): A short free form discription of the actor instance.
            module_functions (dict): A dict of queue names containing an array of module_functions.
            protocol (``wishbone.protocol.Encode``, ``wishbone.protocol.Encode``): A protocol decode or encode instance.
            io_event (bool): When ``True`` Input and Output modules know to expect or emit serialzed wishbone events.
            identification (str): A name assigned to the Wishbone instance, useful for the module to know such as logging.
            disable_exception_handling (bool): If True, exception handling is disabled. Usefull for testing
        '''
        self.name = name
        self.size = size
        self.frequency = frequency
        self.template_functions = template_functions
        self.description = description
        self.module_functions = module_functions
        self.protocol = protocol
        self.io_event = io_event
        self.identification = identification
        self.disable_exception_handling = disable_exception_handling



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  testevent.py
#
#  Copyright 2016 Jelle Smet <development@smetj.net>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from wishbone import Actor
from wishbone.event import Event
from gevent import sleep


class TestEvent(Actor):

    '''**Generates a test event at the chosen interval.**

    The data field of the test event contains the string "test".


    Parameters::

        - interval(float)(1)
           |  The interval in seconds between each generated event.
           |  A value of 0 means as fast as possible.

        - message(string)("test")
           |  The content of the test message.

        - numbered(bool)
           |  When true, appends a sequential number to the end.

        - additional_values(dict)({})
           |  A dictionary of key/value to add to the event.


    Queues::

        - outbox
           |  Contains the generated events.
    '''

    def __init__(self, actor_config, interval=1, message="test", numbered=False, additional_values={}):
        Actor.__init__(self, actor_config)
        self.pool.createQueue("outbox")

    def preHook(self):

        if self.kwargs.interval == 0:
            self.sleep = self.__doNoSleep
        else:
            self.sleep = self.__doSleep

        if self.kwargs.numbered:
            self.generateMessage = self.__doNumber
            self.n = 0
        else:
            self.generateMessage = self.__doNoNumber

        self.sendToBackground(self.produce)

    def produce(self):

        while self.loop():
            message = self.generateMessage(self.kwargs.message)
            event = Event(message)
            for key, value in list(self.kwargs.additional_values.items()):
                event.set(value, key)
            self.submit(event, self.pool.queue.outbox)
            self.sleep()

        self.logging.info("Stopped producing events.")

    def __doSleep(self):
        sleep(self.kwargs.interval)

    def __doNoSleep(self):
        pass

    def __doNumber(self, data):
        self.n += 1
        return "%s %s" % (self.n, data)

    def __doNoNumber(self, data):
        return data
